# -*- coding: utf-8 -*-
"""This module defines functions for comparing normal modes from different
models."""

import numpy as np
from numbers import Integral
from prody import LOGGER, SETTINGS
from prody.utilities import openFile, isListLike

from .nma import NMA
from .modeset import ModeSet
from .mode import Mode, Vector
from .gnm import ZERO
from .analysis import calcFractVariance, calcSqFlucts

__all__ = ['calcOverlap', 'calcCumulOverlap', 'calcSubspaceOverlap', 'calcSpectralOverlap', 
           'calcCovOverlap', 'printOverlapTable', 'writeOverlapTable', 
           'calcSquareInnerProduct','pairModes', 'matchModes', 
           'calcRMSIP', 'calcSIP', 'calcRWSIP']

SO_CACHE = {}
WO_CACHE = {}

def calcOverlap(rows, cols, diag=False):
    """Returns overlap (or correlation) between two sets of modes (*rows* and
    *cols*).  Returns a matrix whose rows correspond to modes passed as *rows*
    argument, and columns correspond to those passed as *cols* argument.
    Both rows and columns are normalized prior to calculating overlap.
    
    This function can now return the diagonal of the overlap matrix if *diag*
    is set to **True**."""

    if not isinstance(rows, (NMA, ModeSet, Mode, Vector, np.ndarray)):
        raise TypeError('rows must be NMA, ModeSet, Mode, Vector, or array, not {0}'
                        .format(type(rows)))
    if not isinstance(cols, (NMA, ModeSet, Mode, Vector, np.ndarray)):
        raise TypeError('cols must be NMA, ModeSet, Mode, or Vector, or array, not {0}'
                        .format(type(cols)))

    if isinstance(rows, np.ndarray):
        num_rows = rows.shape[0]
    else:
        num_rows = rows.numEntries()

    if isinstance(cols, np.ndarray):
        num_cols = cols.shape[0]
    else:
        num_cols = cols.numEntries()

    if num_rows != num_cols:
        raise ValueError('the length of vectors in rows and '
                         'cols must be the same')
    
    if not isinstance(rows, np.ndarray):
        rows = rows.getArray()
    rows *= 1 / (rows ** 2).sum(0) ** 0.5

    if not isinstance(cols, np.ndarray):
        cols = cols.getArray()
    cols *= 1 / (cols ** 2).sum(0) ** 0.5

    if diag:
        overlaps = np.einsum('ij,ji->i', rows.T, cols)
    else:
        overlaps = np.dot(rows.T, cols)
    
    return overlaps


def printOverlapTable(rows, cols):
    """Print table of overlaps (correlations) between two sets of modes.
    *rows* and *cols* are sets of normal modes, and correspond to rows
    and columns of the printed table.  This function may be used to take
    a quick look into mode correspondences between two models.

    >>> # Compare top 3 PCs and slowest 3 ANM modes
    >>> printOverlapTable(p38_pca[:3], p38_anm[:3]) # doctest: +SKIP
    Overlap Table
                            ANM 1p38
                        #1     #2     #3
    PCA p38 xray #1   -0.39  +0.04  -0.71
    PCA p38 xray #2   -0.78  -0.20  +0.22
    PCA p38 xray #3   +0.05  -0.57  +0.06"""

    print(getOverlapTable(rows, cols))


def writeOverlapTable(filename, rows, cols):
    """Write table of overlaps (correlations) between two sets of modes to a
    file.  *rows* and *cols* are sets of normal modes, and correspond to rows
    and columns of the overlap table.  See also :func:`.printOverlapTable`."""

    assert isinstance(filename, str), 'filename must be a string'
    out = openFile(filename, 'w')
    out.write(getOverlapTable(rows, cols))
    out.close()
    return filename


def getOverlapTable(rows, cols):
    """Make a formatted string of overlaps between modes in *rows* and *cols*.
    """

    overlap = calcOverlap(rows, cols)
    if isinstance(rows, Mode):
        rids = [rows.getIndex()]
        rname = str(rows.getModel())
    elif isinstance(rows, NMA):
        rids = np.arange(len(rows))
        rname = str(rows)
    elif isinstance(rows, ModeSet):
        rids = rows.getIndices()
        rname = str(rows.getModel())
    else:
        rids = [0]
        rname = str(rows)
    rlen = len(rids)
    if isinstance(cols, Mode):
        cids = [cols.getIndex()]
        cname = str(cols.getModel())
    elif isinstance(cols, NMA):
        cids = np.arange(len(cols))
        cname = str(cols)
    elif isinstance(cols, ModeSet):
        cids = cols.getIndices()
        cname = str(cols.getModel())
    else:
        cids = [0]
        cname = str(cols)
    clen = len(cids)
    overlap = overlap.reshape((rlen, clen))
    table = 'Overlap Table\n'
    table += (' '*(len(rname)+5) + cname.center(clen*7)).rstrip() + '\n'
    line = ' '*(len(rname)+5)
    for j in range(clen):
        line += ('#{0}'.format(cids[j]+1)).center(7)
    table += line.rstrip() + '\n'
    for i in range(rlen):
        line = rname + (' #{0}'.format(rids[i]+1)).ljust(5)
        for j in range(clen):
            if abs(overlap[i, j]).round(2) == 0.00:
                minplus = ' '
            elif overlap[i, j] < 0:
                minplus = '-'
            else:
                minplus = '+'
            line += (minplus+'{0:-.2f}').format(abs(overlap[i, j])).center(7)
        table += line.rstrip() + '\n'
    return table


def calcCumulOverlap(modes1, modes2, array=False):
    """Returns cumulative overlap of modes in *modes2* with those in *modes1*.
    Returns a number if *modes1* contains a single :class:`.Mode` or a
    :class:`.Vector` instance. If *modes1* contains multiple modes, returns an
    array. Elements of the array correspond to cumulative overlaps for modes
    in *modes1* with those in *modes2*.  If *array* is **True**, returns an array
    of cumulative overlaps. Returned array has the shape ``(len(modes1),
    len(modes2))``.  Each row corresponds to cumulative overlaps calculated for
    modes in *modes1* with those in *modes2*.  Each value in a row corresponds
    to cumulative overlap calculated using up to that many number of modes from
    *modes2*."""

    overlap = calcOverlap(modes1, modes2)
    if array:
        return np.sqrt(np.power(overlap, 2).sum(axis=overlap.ndim-1))
    else:
        return np.sqrt(np.power(overlap, 2).cumsum(axis=overlap.ndim-1))


def calcSubspaceOverlap(modes1, modes2):
    """Returns subspace overlap between two sets of modes (*modes1* and
    *modes2*).  Also known as the root mean square inner product (RMSIP)
    of essential subspaces [AA99]_.  This function returns a single number.

    .. [AA99] Amadei A, Ceruso MA, Di Nola A. On the convergence of the
       conformational coordinates basis set obtained by the essential
       dynamics analysis of proteins' molecular dynamics simulations.
       *Proteins* **1999** 36(4):419-424."""

    overlap = calcOverlap(modes1, modes2)
    if isinstance(modes1, Mode):
        length = 1
    else:
        length = len(modes1)
    rmsip = np.sqrt(np.power(overlap, 2).sum() / length)
    return rmsip

calcRMSIP = calcSubspaceOverlap


def calcRWSIP(modes1, modes2):
    """Returns root weighted square inner product (RWSIP)
    of essential subspaces [VC07]_.  This function returns a single number.

    .. [VC07] Carnevale V, Pontiggia F, Micheletti C. Structural and dynamical 
       alignment of enzymes with partial structural similarity.
       *J Phys Condens Matter.* **2007** 19:285206."""

    overlap = calcOverlap(modes1, modes2)
    if not isinstance(overlap, np.ndarray):
        overlap = np.array(overlap)
    
    if isinstance(modes1, Mode):
        length1 = 1
    else:
        length1 = len(modes1)

    if isinstance(modes2, Mode):
        length2 = 1
    else:
        length2 = len(modes2)

    vars1 = modes1.getVariances()
    if length1 == 1:
        vars1 = [vars1]
        overlap = overlap.reshape(-1, 1)

    vars2 = modes2.getVariances()
    if length2 == 1:
        vars2 = [vars2]
        overlap = overlap.reshape(1, -1)

    numerator = np.sum([vars1[l] * vars2[m] * overlap[l, m]**2
                        for l in range(length1)
                        for m in range(length2)])
    
    denominator = np.dot(vars1[:length1], vars2[:length1])
    
    rwsip = np.sqrt(numerator/denominator)
    return rwsip


def calcSquareInnerProduct(modes1, modes2):
    """Returns the square inner product (SIP) of fluctuations [SK02]_.  
    This function returns a single number.

    .. [SK02] Kundu S, Melton JS, Sorensen DC, Phillips GN: Dynamics of 
        proteins in crystals: comparison of experiment with simple models. 
        *Biophys J.* **2002**, 83:723-732.
        
    """
    if isinstance(modes1, (NMA, ModeSet)):
        w1 = calcSqFlucts(modes1)
    elif isListLike(modes1):
        w1 = modes1
    else:
        raise TypeError('modes1 should be a profile or an NMA or ModeSet object')

    if isinstance(modes2, (NMA, ModeSet)):
        w2 = calcSqFlucts(modes2)
    elif isListLike(modes2):
        w2 = modes2
    else:
        raise TypeError('modes2 should be a profile or an NMA or ModeSet object')

    return np.dot(w1, w2)**2 / (np.dot(w1, w1) * np.dot(w2, w2))

calcSIP = calcSquareInnerProduct


def calcSpectralOverlap(modes1, modes2, weighted=False, turbo=False):
    """Returns overlap between covariances of *modes1* and *modes2*.  Overlap
    between covariances are calculated using normal modes (eigenvectors),
    hence modes in both models must have been calculated.  This function
    implements equation 11 in [BH02]_.

    .. [BH02] Hess B. Convergence of sampling in protein simulations.
        *Phys Rev E* **2002** 65(3):031910.
    
    :arg weighted: if **True** then covariances are weighted by the trace.
    :type weighted: bool
    """

    if modes1.is3d() ^ modes2.is3d():
        raise TypeError('models must be either both 1-dimensional or 3-dimensional')
    if modes1.numAtoms() != modes2.numAtoms():
        raise ValueError('modes1 and modes2 must have same number of atoms')

    if isinstance(modes1, Mode):
        if weighted:
            varA = np.array([calcFractVariance(modes1)])
        else:
            varA = np.array([modes1.getVariance()])
        I = np.array([modes1.getIndex()])
    else:
        if weighted:
            varA = calcFractVariance(modes1)
        else:
            varA = modes1.getVariances()

        try:
            I = modes1.getIndices()
        except:
            try:
                modes1 = modes1[:]
                I = modes1.getIndices()
            except:
                raise TypeError('modes1 should be ModeSet or an object from which a ModeSet can be obtained')

    if isinstance(modes2, Mode):
        if weighted:
            varB = np.array([calcFractVariance(modes2)])
        else:
            varB = np.array([modes2.getVariance()])
        J = np.array([modes2.getIndex()])
    else:
        if weighted:
            varB = calcFractVariance(modes2)
        else:
            varB = modes2.getVariances()

        try:
            J = modes2.getIndices()
        except:
            try:
                modes2 = modes2[:]
                J = modes2.getIndices()
            except:
                raise TypeError('modes2 should be ModeSet or an object from which a ModeSet can be obtained')

    if turbo:
        model1 = modes1.getModel()
        model2 = modes2.getModel()
        if weighted:
            CACHE = WO_CACHE
        else:
            CACHE = SO_CACHE
        if (model1, model2) in CACHE:
            weights = CACHE[(model1, model2)]
        elif (model2, model1) in CACHE:
            weights = CACHE[(model2, model1)]
        else:
            farrayA = model1._getArray()
            farrayB = model2._getArray()

            fvarA = model1.getVariances()
            fvarB = model2.getVariances()

            dotAB = np.dot(farrayA.T, farrayB)**2
            outerAB = np.outer(fvarA**0.5, fvarB**0.5)
            CACHE[(model1, model2)] = weights = outerAB * dotAB
        
        weights = weights[I, :][:, J]
    else:
        arrayA = modes1._getArray()
        arrayB = modes2._getArray()

        dotAB = np.dot(arrayA.T, arrayB)**2
        outerAB = np.outer(varA**0.5, varB**0.5)
        weights = outerAB * dotAB

    diff = (np.sum(varA.sum() + varB.sum()) - 2 * np.sum(weights))

    if diff < ZERO:
        diff = 0
    else:
        diff = diff ** 0.5
    return 1 - diff / np.sqrt(varA.sum() + varB.sum())

def calcCovOverlap(modes1, modes2, turbo=False):
    """Returns overlap between covariances of *modes1* and *modes2*.  Overlap
    between covariances are calculated using normal modes (eigenvectors),
    hence modes in both models must have been calculated.  This function
    implements equation 11 in [BH02]_."""
    return calcSpectralOverlap(modes1, modes2, turbo=turbo)

def pairModes(modes1, modes2, **kwargs):
    """Returns the optimal matches between *modes1* and *modes2*. *modes1* 
    and *modes2* should have equal number of modes, and the function will 
    return a nested list where each item is a list containing a pair of modes.

    :arg index: if **True** then indices of modes will be returned instead of 
        :class:`Mode` instances.
    :type index: bool
    """

    index = kwargs.pop('index', False)
    method = kwargs.pop('method', None)

    if method is None:
        from scipy.optimize import linear_sum_assignment
        method = linear_sum_assignment

    if not (isinstance(modes1, (ModeSet, NMA)) \
        and isinstance(modes2, (ModeSet, NMA))):
        raise TypeError('modes1 and modes2 should be ModeSet or NMA instances')

    if len(modes1) != len(modes2):
        raise ValueError('the same number of modes should be provided')
    overlaps = calcOverlap(modes1, modes2)

    costs = 1 - abs(overlaps)
    row_ind, col_ind = method(costs)

    if index:
        return row_ind, col_ind

    if isinstance(modes1, ModeSet):
        row_ind = modes1._indices[row_ind]

    if isinstance(modes2, ModeSet):
        col_ind = modes2._indices[col_ind]

    outmodes1 = ModeSet(modes1.getModel(), row_ind)
    outmodes2 = ModeSet(modes2.getModel(), col_ind)

    return outmodes1, outmodes2

def _pairModes_wrapper(args):
    modeset0, modesets, index = args

    ret = []
    for modeset in modesets:
        _, reordered_modeset = pairModes(modeset0, modeset, index=index)
        ret.append(reordered_modeset)
    return ret

def matchModes(*modesets, **kwargs):
    """Returns the matches of modes among *modesets*. Note that the first 
    modeset will be treated as the reference so that only the matching 
    of each modeset to the first modeset is guaranteed to be optimal.
    
    :arg index: if **True** then indices of modes will be returned instead of 
                :class:`Mode` instances
    :type index: bool

    :arg turbo: if **True** then the computation will be performed in parallel. 
                The number of threads is set to be the same as the number of 
                CPUs. Assigning a number will specify the number of threads to be 
                used. Note that if writing a script, ``if __name__ == '__main__'`` 
                is necessary to protect your code when multi-tasking. 
                See https://docs.python.org/2/library/multiprocessing.html for details.
                Default is **False**
    :type turbo: bool, int
    """

    index = kwargs.pop('index', False)
    turbo = kwargs.pop('turbo', False)

    n_worker = None
    if not isinstance(turbo, bool):
        try:
            n_worker = int(turbo)
        except TypeError:
            raise TypeError('turbo should be Boolean or a number')

    modeset0 = modesets[0]
    if index:
        ret = [modeset0.getIndices()]
    else:
        ret = [modeset0]

    n_modes = len(modeset0)
    n_sets = len(modesets)
    if n_sets == 1:
        return ret
    elif n_sets == 0:
        raise ValueError('at least one modeset should be given')

    if turbo:
        from multiprocessing import Pool, cpu_count
        from math import ceil
        
        if not n_worker:
            n_worker = cpu_count()

        LOGGER.info('Matching {0} modes across {1} modesets with {2} threads...'
                        .format(n_modes, n_sets, n_worker))

        pool = Pool(n_worker)
        n_sets_per_worker = ceil((n_sets - 1) / n_worker)
        args = []
        for i in range(n_worker):
            start = i*n_sets_per_worker + 1
            end = (i+1)*n_sets_per_worker + 1
            subset = modesets[start:end]
            args.append((modeset0, subset, index))
        nested_ret = pool.map(_pairModes_wrapper, args)
        for entry in nested_ret:
            ret.extend(entry)

        pool.close()
        pool.join()
    else:
        LOGGER.progress('Matching {0} modes across {1} modesets...'
                        .format(n_modes, n_sets), n_sets, '_prody_matchModes')
        for i, modeset in enumerate(modesets):
            LOGGER.update(i, label='_prody_matchModes')
            if i > 0:
                _, reordered_modeset = pairModes(modeset0, modeset, index=index, **kwargs)
                ret.append(reordered_modeset)
        LOGGER.finish()
    
    return ret



# -*- coding: utf-8 -*-
"""This module defines classes for handling mode data."""

import numpy as np

__all__ = ['Mode', 'Vector']


class VectorBase(object):
    """A base class for :class:`Mode` and :class:`Vector`.

    This base class defines some shared methods, such as scalar multiplication
    or addition of mode instances.

    Defined operations are:

        * Absolute value (abs(mode)) returns mode length
        * Additive inverse (-mode)
        * Mode addition (mode1 + mode2)
        * Mode subtraction (mode1 - mode2)
        * Scalar multiplication (x*mode or mode*x)
        * Division by a scalar (mode/x)
        * Dot product (mode1*mode2)
        * Power (mode**x)"""

    __slots__ = []

    def __abs__(self):

        return np.sqrt((self._getArray()**2).sum())

    def __neg__(self):

        return Vector(-self._getArray(), '-({0})'.format(str(self)),
                      self.is3d())

    def __div__(self, other):

        try:
            result = self._getArray() / other
        except Exception as err:
            raise TypeError('{0} is not a scalar {1}'
                            .format(other, str(err)))
        return Vector(result, '({0})/{1}'.format(str(self), other),
                      self.is3d())

    def __idiv__(self, other):

        return self.__div__(other)

    def __mul__(self, other):
        """Returns scaled mode or dot product between modes."""

        try:
            other = other._getArray()
        except AttributeError:
            try:
                result = other * self._getArray()
            except Exception as err:
                raise TypeError('{0} is not a scalar or a mode ({1})'
                                .format(other, str(err)))
            else:
                return Vector(result, '({1})*{0}'.format(other, str(self)),
                              self.is3d())
        else:
            try:
                return np.dot(self._getArray(), other)
            except Exception:
                raise ValueError('{0} and {1} do not have same dimensions '
                                 '({2})'.format(str(self), str(other),
                                                str(err)))

    def __rmul__(self, other):
        """Returns scaled mode or dot product between modes."""

        try:
            other = other._getArray()
        except AttributeError:
            try:
                result = other * self._getArray()
            except Exception as err:
                raise TypeError('{0} is not a scalar or a mode ({1})'
                                .format(other, str(err)))
            else:
                return Vector(result, '{0}*({1})'.format(other, str(self)),
                              self.is3d())
        else:
            try:
                return np.dot(self._getArray(), other)
            except Exception:
                raise ValueError('{0} and {1} do not have same dimensions '
                                 '({2})'.format(str(self), str(other),
                                                str(err)))

    def __imul__(self, other):

        return self.__mul__(other)

    def __add__(self, other):

        if isinstance(other, VectorBase):
            if len(self) != len(other):
                raise ValueError('modes do not have the same length')
            return Vector(self._getArray() + other._getArray(),
                          '({0}) + ({1})'.format(str(self), str(other)),
                          self.is3d())
        else:
            raise TypeError('{0} is not a mode instance'.format(other))

    def __radd__(self, other):

        if isinstance(other, VectorBase):
            if len(self) != len(other):
                raise ValueError('modes do not have the same length')
            return Vector(self._getArray() + other._getArray(),
                          '({0}) + ({1})'.format(str(other), str(self)),
                          self.is3d())
        else:
            raise TypeError('{0} is not a mode instance'.format(other))

    def __iadd__(self, other):

        return self.__add__(other)

    def __sub__(self, other):

        if isinstance(other, VectorBase):
            if len(self) != len(other):
                raise ValueError('modes do not have the same length')
            return Vector(self._getArray() - other._getArray(),
                          '({0}) - ({1})'.format(str(self), str(other)),
                          self.is3d())
        else:
            raise TypeError('{0} is not a mode instance'.format(other))

    def __rsub__(self, other):

        if isinstance(other, VectorBase):
            if len(self) != len(other):
                raise ValueError('modes do not have the same length')
            return Vector(other._getArray() - self._getArray(),
                          '({0}) - ({1})'.format(str(other), str(self)),
                          self.is3d())
        else:
            raise TypeError('{0} is not a mode instance'.format(other))

    def __isub__(self, other):

        return self.__sub__(other)

    def __pow__(self, other):

        try:
            result = self._getArray() ** other
        except Exception as err:
            raise TypeError('{0} is not a scalar ({0})'
                            .format(other, str(err)))
        else:
            return Vector(result, '({0})**{1}'.format(str(self), other),
                          self.is3d())

    def getArray(self):
        """Returns a copy of array."""

        pass

    def _getArray(self):
        """Returns array."""

        pass

    def numAtoms(self):
        """Returns number of atoms."""

        pass

    def is3d(self):
        """Returns **True** if vector is 3d."""

        pass

    def getArrayNx3(self):
        """Returns a copy of array with shape (N, 3)."""

        if self.is3d():
            return self.getArray().reshape((self.numAtoms(), 3))
        else:
            return self.getArray()

    def _getArrayNx3(self):
        """Returns a copy of array with shape (N, 3)."""

        if self.is3d():
            return self._getArray().reshape((self.numAtoms(), 3))
        else:
            return self._getArray()

    def numModes(self):
        """Returns 1."""

        return 1


class Mode(VectorBase):

    """A class to provide access to and operations on mode data.

    """

    __slots__ = ['_model', '_index']

    def __init__(self, model, index):
        """Initialize mode object as part of an NMA model.

        :arg model: a normal mode analysis instance
        :type model: :class:`.NMA`, :class:`.GNM`, or :class:`.PCA`
        :arg index: index of the mode
        :type index: int"""

        self._model = model
        self._index = int(index)

    def __len__(self):

        return self.numEntries()

    def __repr__(self):

        return '<Mode: {0} from {1}>'.format(self._index + 1,
                                             str(self._model))

    def __str__(self):

        return 'Mode {0} from {1}'.format(self._index+1, str(self._model))

    def __int__(self):

        return self._index

    def __float__(self):

        return self.getEigval()

    def is3d(self):
        """Returns **True** if mode instance is from a 3-dimensional model."""

        return self._model.is3d()

    def numAtoms(self):
        """Returns number of atoms."""

        return self._model.numAtoms()

    def numDOF(self):
        """Returns number of degrees of freedom."""

        return self._model.numDOF()

    def numEntries(self):
        """Returns number of entries in the eigenvector."""

        return len(self._getArray())

    def getTitle(self):
        """A descriptive title for the mode instance."""

        return str(self)

    def getIndex(self):
        """Returns the index of the mode.  Note that mode indices are
        zero-based."""

        return self._index

    def getModel(self):
        """Returns the model that the mode instance belongs to."""

        return self._model

    def getArray(self):
        """Returns a copy of the normal mode array (eigenvector)."""

        return self._model.getArray()[:, self._index].copy()

    getEigvec = getArray
    getEigvecs = getEigvec

    def _getArray(self):
        """Returns a copy of the normal mode array (eigenvector)."""

        return self._model._getArray()[:, self._index]

    def getEigval(self):
        """Returns normal mode eigenvalue.  For :class:`.PCA` and :class:`.EDA`
        models built using coordinate data in Å, unit of eigenvalues is |A2|.
        For :class:`.ANM` and :class:`.GNM`, on the other hand, eigenvalues
        are in arbitrary or relative units but they correlate with stiffness
        of the motion along associated eigenvector."""

        return self._model._eigvals[self._index]

    getEigvals = getEigval

    def getVariance(self):
        """Returns variance along the mode.  For :class:`.PCA` and :class:`.EDA`
        models built using coordinate data in Å, unit of variance is |A2|.  For
        :class:`.ANM` and :class:`.GNM`, on the other hand, variance is the
        inverse of the eigenvalue, so it has arbitrary or relative units."""

        return self._model._vars[self._index]
    
    getVariances = getVariance

class Vector(VectorBase):

    """A class to provide operations on a modified mode array.  This class
    holds only mode array (i.e. eigenvector) data, and has no associations
    with an NMA instance.  Scalar multiplication of :class:`Mode` instance
    or addition of two :class:`Mode` instances results in a :class:`Vector`
    instance."""

    __slots__ = ['_title', '_array', '_is3d', '_n_atoms']

    def __init__(self, array, title='Unknown', is3d=True):
        """Instantiate with a name, an array, and a 3d flag."""

        try:
            ndim, shape = array.ndim, array.shape
        except AttributeError:
            array = np.array(array)
            ndim, shape = array.ndim, array.shape
        if ndim != 1:
            raise ValueError('array.ndim must be 1')
        is3d = bool(is3d)
        if is3d and shape[0] % 3 != 0:
            raise ValueError('len(array) must be a multiple of 3')
        self._title = str(title)
        self._array = array
        self._is3d = is3d

        if self._is3d:
            self._n_atoms = len(self._array)//3
        else:
            self._n_atoms = len(self._array)

    def __len__(self):
        return len(self._array)

    def __repr__(self):
        return '<Vector: {0}>'.format(self._title)

    def __str__(self):
        return self._title

    def is3d(self):
        """Returns **True** if vector instance describes a 3-dimensional
        property, such as a deformation for a set of atoms."""

        return self._is3d

    def getTitle(self):
        """Get the descriptive title for the vector instance."""

        return self._title

    def setTitle(self, title):
        """Set the descriptive title for the vector instance."""

        self._title = str(title)

    def getArray(self):
        """Returns a copy of array."""

        return self._array.copy()

    def _getArray(self):
        """Returns array."""

        return self._array

    def getNormed(self):
        """Returns mode after normalizing it."""

        return Vector(self._array/(self._array**2).sum()**0.5,
                      '({0})/||{0}||'.format(self._title), self._is3d)

    def numDOF(self):
        """Returns number of degrees of freedom."""

        return len(self._array)

    def numEntries(self):
        """Returns number of entries in the vector."""

        return len(self._array)

    def numAtoms(self):
        """Returns number of atoms.  For a 3-dimensional vector, returns length
        of the vector divided by 3."""

        return self._n_atoms



"""This module contains unit tests for :mod:`~prody.ensemble`."""

from os.path import join
from prody.tests import TestCase

from numpy.testing import assert_equal, assert_allclose

from prody import DCDFile, writeDCD, parseDCD

from prody.tests import TEMPDIR
from prody.tests.ensemble import ALLATOMS, ENSEMBLE, RTOL, ATOL, DCD

class TestDCDFile(TestCase):

    def setUp(self):

        self.dcd = join(TEMPDIR, 'temp.dcd')

    def testWriteDCD(self):
        dcd = writeDCD(self.dcd, ALLATOMS)
        self.assertEqual(dcd, self.dcd, 'failed to write DCD file')

    def testParseDCD(self):
        e = parseDCD(writeDCD(self.dcd, ALLATOMS))
        assert_equal(e._getCoordsets(), DCD._getCoordsets(),
                     err_msg='failed to parse DCD file correctly')

    def testWrite(self):
        dcd = DCDFile(self.dcd, 'w')
        dcd.write(ENSEMBLE.getCoordsets())
        dcd.close()
        e = parseDCD(self.dcd)
        assert_allclose(e._getCoordsets(), ENSEMBLE._getCoordsets(),
                        rtol=RTOL, atol=ATOL,
                        err_msg='failed to parse DCD file correctly')

    def testWriteModeAppend(self):
        dcd = DCDFile(writeDCD(self.dcd, ENSEMBLE), 'a')
        dcd.write(ENSEMBLE.getCoordsets())
        dcd.close()
        e = parseDCD(self.dcd)
        n_csets = len(ENSEMBLE)
        coordsets = e._getCoordsets()
        assert_equal(coordsets, coordsets,
                     'failed to parse DCD file correctly')
        assert_allclose(coordsets[:n_csets], ENSEMBLE._getCoordsets(),
                        rtol=RTOL, atol=ATOL,
                        err_msg='failed to parse DCD file correctly')



# -*- coding: utf-8 -*-
"""This module defines atom flags that are used in :ref:`selections`.
You can read this page in interactive sessions using ``help(flags)``.

Flag labels can be used in atom selections:

.. ipython:: python

   from prody import *
   p = parsePDB('1ubi')
   p.select('protein')

Flag labels can be combined with dot operator as follows to make selections:

.. ipython:: python

   p.protein
   p.protein.acidic  # selects acidic residues

Flag labels can be prefixed with ``'is'`` to check whether all atoms in
an :class:`.Atomic` instance are flagged the same way:

.. ipython:: python

   p.protein.ishetero
   p.water.ishetero

Flag labels can also be used to make quick atom counts:

.. ipython:: python

   p.numAtoms()
   p.numAtoms('protein')
   p.numAtoms('water')"""

from re import compile as recompile
from time import time
from collections import defaultdict

from numpy import array, ones, zeros

from prody import SETTINGS, LOGGER
from prody.utilities import openData
from prody.utilities import joinLinks, joinTerms, wrapText

__all__ = ['flagDefinition', 'listNonstdAAProps', 'getNonstdProperties',
           'addNonstdAminoacid', 'delNonstdAminoacid', 'NAMAP']


TIMESTAMP_KEY = 'flags_timestamp'
DEFINITIONS_KEY = 'flags_definitions'
NONSTANDARD_KEY = 'flags_nonstandard'

ALIASES = {}
PLANTERS = {}
EDITORS = {}
FIELDS = defaultdict(set)  # flags that fill be nulled when a field changes
FIELDSDEFAULT = ['name', 'resname', 'resnum']
TIMESTAMP = 0

PDBLIGSUM = ('.. _{0}: '
             'http://www.pdb.org/pdb/ligand/ligandsummary.do?hetId={0}\n')

STANDARDAA = ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY', 'HIS',
              'ILE', 'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER', 'THR', 'TRP',
              'TYR', 'VAL']

NONSTANDARD = {
    'ASX': set(['acyclic', 'surface', 'polar', 'medium']),
    'GLX': set(['acyclic', 'surface', 'large', 'polar']),
    'ASH': set(['acyclic', 'acidic', 'surface', 'polar', 'medium']),
    'GLH': set(['acyclic', 'acidic', 'surface', 'large', 'polar']),
    'CSO': set(['acyclic', 'neutral', 'surface', 'medium', 'polar']),
    'CYX': set(['acyclic', 'neutral', 'buried', 'medium', 'polar']),
    'HIP': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HID': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HIE': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HSD': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HSE': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HSP': set(['cyclic', 'acidic', 'surface', 'large', 'polar']),
    'HISD': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HISE': set(['cyclic', 'basic', 'surface', 'large', 'polar']),
    'HISP': set(['cyclic', 'acidic', 'surface', 'large', 'polar']),
    'LYN': set(['acyclic', 'neutral', 'surface', 'large', 'polar']),
    'TYM': set(['cyclic', 'aromatic', 'surface', 'basic', 'large', 'polar']),
    'ARN': set(['acyclic', 'neutral', 'surface', 'large', 'polar']),
    'MSE': set(['acyclic', 'neutral', 'buried', 'large']),
    'CME': set(['acyclic', 'neutral', 'buried', 'large']),
    'SEC': set(['acyclic', 'neutral', 'buried', 'polar', 'medium']),
    'SEP': set(['acyclic', 'surface', 'acidic', 'large', 'polar']),
    'TPO': set(['acyclic', 'surface', 'acidic', 'large', 'polar']),
    'PTR': set(['cyclic', 'aromatic', 'surface', 'acidic', 'large', 'polar']),
    'PHD': set(['acyclic', 'acidic', 'surface', 'polar', 'large']),
    'XLE': set(['aliphatic', 'acyclic', 'buried', 'hydrophobic', 'large']),
    'XAA': set(),
    'MEN': set(),
    'CSB': set()
}

DEFAULTS = {

    'stdaa': set(['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY',
                  'HIS', 'ILE', 'LEU', 'LYS', 'MET', 'PHE', 'PRO', 'SER',
                  'THR', 'TRP', 'TYR', 'VAL']),

    'nonstdaa': set(NONSTANDARD),

    # GLY is excluded from this list
    'aliphatic': set(['ALA', 'ILE', 'LEU', 'VAL', 'PRO', 'GLY']),
    'aromatic': set(['HIS', 'PHE', 'TRP', 'TYR']),

    'acidic': set(['ASP', 'GLU']),
    'basic': set(['LYS', 'ARG', 'HIS']),
    'neutral': set(['ALA', 'ASN', 'CYS', 'GLN', 'GLY', 'ILE', 'LEU', 'MET',
                    'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL']),

    'buried': set(['ALA', 'CYS', 'ILE', 'LEU', 'MET', 'PHE', 'TRP', 'VAL']),
    'surface': set(['ARG', 'ASN', 'ASP', 'GLN', 'GLU', 'GLY', 'HIS', 'LYS',
                    'PRO', 'SER', 'THR', 'TYR']),

    'cyclic': set(['HIS', 'PHE', 'PRO', 'TRP', 'TYR']),
    'acyclic': set(['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY',
                    'ILE', 'LEU', 'LYS', 'MET', 'SER', 'THR', 'VAL']),

    'hydrophobic': set(['ALA', 'ILE', 'LEU', 'MET', 'PHE', 'PRO', 'TRP',
                        'VAL']),
    'polar': set(['ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY', 'HIS',
                  'LYS', 'SER', 'THR', 'TYR']),

    'small': set(['ALA', 'GLY', 'SER']),
    'medium': set(['ASN', 'ASP', 'CYS', 'PRO', 'THR', 'VAL']),
    'large': set(['ARG', 'GLN', 'GLU', 'HIS', 'ILE', 'LEU', 'LYS', 'MET',
                  'PHE', 'TRP', 'TYR']),


    'hydrogen': '[0-9]?H.*',
    'carbon': 'C.*',
    'nitrogen': 'N.*',
    'oxygen': 'O.*',
    'sulfur': 'S.*',

    'nucleobase': set(['GUN', 'ADE', 'CYT', 'THY', 'URA']),
    'nucleotide': set(['DA', 'DC', 'DG', 'DT', 'DU', 'A', 'C', 'G', 'T', 'U']),
    'nucleoside': set(['ADN', 'AMP', 'ADP', 'ATP',
                       'CMP', # cyclic AMP
                       'A2P', 'A3P', # multi site diphosphates
                       'AGS', # ATP-gamma-S
                       'CTN', 'C5P', 'CDP', 'CTP',
                       'C2P', 'C3P', # other site monophosphates
                       'GMP', '5GP', 'GDP', 'GTP',
                       'THM', 'TMP', 'TPP', 'TTP',
                       'URI', 'UMP', 'UDP', 'UTP']),

    'at': set(['ADE', 'A', 'THY', 'T']),
    'cg': set(['CYT', 'C', 'GUN', 'G']),
    'purine': set(['ADE', 'A', 'GUN', 'G']),
    'pyrimidine': set(['CYT', 'C', 'THY', 'T', 'URA', 'U']),

    'water': set(['HOH', 'DOD', 'WAT', 'TIP3', 'H2O', 'OH2', 'TIP', 'TIP2',
                  'TIP4', 'SPC', 'SOL', 'T3P']),

    'ion': set(['AL', 'BA', 'CA', 'CD', 'CL', 'CO', 'CS', 'CU', 'CU1', 'CUA',
                'HG', 'IN', 'IOD', 'K', 'MG', 'MN', 'MN3', 'NA', 'PB', 'PT', 'RB',
                'TB', 'TL', 'WO4', 'YB', 'ZN']),
    'ion_other': set(['CAL', 'CES', 'CLA', 'POT', 'SOD', 'ZN2', 'CU2', 'CU2P']),


    'lipid': set(['GPE', 'LPP', 'OLA', 'SDS', 'STE']),
    'lipid_other': set(['DLPE', 'DMPC', 'LPPC', 'OLEO', 'PALM', 'PCGL', 'POPC',
                        'POPE', 'STEA']),

    'sugar': set(['GLC', 'GLO', 'BGC']),
    'sugar_other': set(['AGLC']),

    'heme': set(['1FH', '2FH', 'DDH', 'DHE', 'HAS', 'HDD', 'HDE', 'HDM',
                 'HEA', 'HEB', 'HEC', 'HEM', 'HEO', 'HES', 'HEV', 'NTE',
                 'SRM', 'VER']),
    'heme_other': set(['HEMO', 'HEMR'])

}

DEFAULTS['backbone'] = DEFAULTS['bb'] = set(['CA', 'C', 'O', 'N'])
DEFAULTS['backbonefull'] = DEFAULTS['bbfull'] = set(['CA', 'C', 'O', 'N', 'H',
                                                     'H1', 'H2', 'H3', 'OXT'])

CATEGORIES = {
    'aromaticity': set(['aromatic', 'aliphatic']),
    'cyclic': set(['acyclic', 'cyclic']),
    'charge': set(['acidic', 'basic', 'neutral']),
    'depth': set(['buried', 'surface']),
    'hydrophobicity': set(['hydrophobic', 'polar']),
    'size': set(['small', 'medium', 'large']),
}

CATEGORIZED = {}
for catgroup in CATEGORIES.values():
    for category in catgroup:
        CATEGORIZED[category] = list(DEFAULTS[category])
INDEPENDENT = list(CATEGORIZED)
CATEGORIZED['charged'] = list(CATEGORIZED['acidic'])
CATEGORIZED['charged'].extend(CATEGORIZED['basic'])

for resi, cats in NONSTANDARD.items():
    for cat in cats:
        CATEGORIZED[cat].append(resi)


def addPlanter(func, *labels, **kwargs):

    aliases = kwargs.get('aliases', True)
    editor = kwargs.get('editor', False)
    for label in labels:
        PLANTERS[label] = func
        if aliases:
            ALIASES[label] = labels
        else:
            ALIASES[label] = [label]
        if editor:
            EDITORS[label] = editor
    for field in kwargs.get('fields', FIELDSDEFAULT):
        FIELDS[field].update(labels)


def updateNonstandard(nonstd):

    SETTINGS[NONSTANDARD_KEY] = nonstd
    SETTINGS[TIMESTAMP_KEY] = int(time())
    SETTINGS.save()
    updateDefinitions()


def changeDefinitions(**kwargs):

    defs = SETTINGS.get(DEFINITIONS_KEY, {})
    defs.update(kwargs)
    SETTINGS[DEFINITIONS_KEY] = defs
    SETTINGS[TIMESTAMP_KEY] = int(time())
    SETTINGS.save()
    updateDefinitions()


def resetDefinitions(flag):

    if flag == 'all':
        SETTINGS.pop(DEFINITIONS_KEY, None)
        SETTINGS.pop(NONSTANDARD_KEY, None)
        SETTINGS[TIMESTAMP_KEY] = int(time())
        SETTINGS.save()
        updateDefinitions()
    elif flag == 'nonstdaa':
        SETTINGS.pop(NONSTANDARD_KEY, None)
        SETTINGS[TIMESTAMP_KEY] = int(time())
        SETTINGS.save()
        updateDefinitions()
    else:
        try:
            SETTINGS.pop(DEFINITIONS_KEY, {}).pop(flag)
        except KeyError:
            pass
        else:
            SETTINGS[TIMESTAMP_KEY] = int(time())
            SETTINGS.save()
            updateDefinitions()


def changeResnames(flag, resnames):
    """Change the list of residue names associated with a flag.  *flag* must
    be a string, and *resnames* may be a list, tuple, or set of strings.  The
    existing list of residue names will be overwritten with the given residue
    names."""

    if flag not in EDITORS:
        raise ValueError('{0} is not an editable flag'.format(repr(flag)))

    resnames = [str(rn) for rn in resnames]
    changeDefinitions(**{flag: resnames})


def changeNameRegex(flag, regex):
    """Set regular expression used for flagging elements based on atom names.
    See :ref:`element-flags` for the default list of flags."""

    if flag not in EDITORS:
        raise ValueError('{0} is not an editable flag'.format(repr(flag)))

    try:
        recompile(regex)
    except Exception as err:
        raise ValueError('{0} is not a valid regular expression, {1}'
                         .format(repr(regex), str(err)))
    else:
        changeDefinitions(**{flag: regex})


def changeBackbone(flag, names):
    """Set protein :term:`backbone` or :term:`backbonefull` atom names.
    *names* must be a list of atom names."""

    if flag not in EDITORS:
        raise ValueError('{0} is not an editable flag'.format(repr(flag)))

    names = [str(nm) for nm in names]

    if flag.endswith('full'):
        changeDefinitions(bbfull=names, backbonefull=names)
    else:
        changeDefinitions(bb=names, backbone=names)


__doc__ += """

Protein
-------------------------------------------------------------------------------

.. glossary::

   protein
   aminoacid
      indicates the twenty standard amino acids (:term:`stdaa`) and some
      non-standard amino acids (:term:`nonstdaa`) described below.  Residue
      must also have an atom named ``'CA'`` in addition to having a qualifying
      residue name.


   stdaa
      {stdaa}


   nonstdaa
      indicates one of the following residues:

      ==========  ===================================================
      `ASX`_ (B)  asparagine or aspartic acid
      `GLX`_ (Z)  glutamine or glutamic acid
      `CSO`_ (C)  S-hydroxycysteine
      `HIP`_ (H)  ND1-phosphohistidine
       HSD   (H)  prototropic tautomer of histidine, H on ND1 (CHARMM)
       HSE   (H)  prototropic tautomer of histidine, H on NE2 (CHARMM)
       HSP   (H)  protonated histidine
      `MSE`_      selenomethionine
      `SEC`_ (U)  selenocysteine
      `SEP`_ (S)  phosphoserine
      `TPO`_ (T)  phosphothreonine
      `PTR`_ (Y)  O-phosphotyrosine
       XLE   (J)  leucine or isoleucine
       XAA   (X)  unspecified or unknown
      ==========  ===================================================

      You can modify the list of non-standard amino acids using
      :func:`addNonstdAminoacid`, :func:`delNonstdAminoacid`, and
      :func:`listNonstdAAProps`.


   calpha
   ca
      Cα atoms of :term:`protein` residues, same as selection
      ``'name CA and protein'``


   backbone
   bb
      non-hydrogen backbone atoms of :term:`protein` residues, same as
      selection ``'name CA C O N and protein'``


   backbonefull
   bbfull
      backbone atoms of :term:`protein` residues, same as selection
      ``'name CA C O N H H1 H2 H3 OXT and protein'``


   sidechain
   sc
      side-chain atoms of :term:`protein` residues, same as selection
      ``'protein and not backbonefull'``


""".format(

stdaa = wrapText('indicates the standard amino acid residues: ' +
                 joinLinks(DEFAULTS['stdaa'], last=', and ', sort=True),
                 subsequent_indent=' '*6))


cats = list(CATEGORIZED)
cats.sort()

for cat in cats:
    res = CATEGORIZED[cat]
    res.sort()
    __doc__ += """
   {cat}
       {res}

""".format(cat=cat, res=wrapText('residues ' + ', '.join(res),
                                 subsequent_indent=' '*6))

for resi in DEFAULTS['stdaa']:
    __doc__ += PDBLIGSUM.format(resi)
for resi in DEFAULTS['nonstdaa']:
    __doc__ += PDBLIGSUM.format(resi)


__doc__ += """

Nucleic
-------------------------------------------------------------------------------

.. glossary::

   nucleic
      indicates :term:`nucleobase`, :term:`nucleotide`, and some
      :term:`nucleoside` derivatives that are described below, so it is same
      as ``'nucleobase or nucleotide or nucleoside'``.

   nucleobase
      indicates `ADE`_ (adenine), `GUN`_ (guanine), `CYT`_
      (cytosine), `THY`_ (thymine), and `URA`_ (uracil).

   nucleotide
      indicates residues with the following names:

      =======  ==================================
      `DA`_    2'-deoxyadenosine-5'-monophosphate
      `DC`_    2'-deoxycytidine-5'-monophosphate
      `DG`_    2'-deoxyguanosine-5'-monophosphate
      `DT`_    2'-deoxythymidine-5'-monophosphate
      `DU`_    2'-deoxyuridine-5'-monophosphate
      `A`_     adenosine-5'-monophosphate
      `C`_     cytidine-5'-monophosphate
      `G`_     guanosine-5'-monophosphate
      `T`_     2'-deoxythymidine-5'-monophosphate
      `U`_     uridine-5'-monophosphate
      =======  ==================================

   nucleoside
      indicates following nucleosides and their derivatives that are recognized by *PDB*:

      =======  ================================
      `ADN`_   adenosine
      `AMP`_   adenosine-5'-monophosphate
      `ADP`_   adenosine-5'-diphosphate
      `ATP`_   adenosine-5'-triphosphate
      `AGS`_   adenosine-5'-triphosphate-gamma-S
      `CMP`_   cyclic adenosine-3',5'-monophosphate
      `A2P`_   adenosine-2',5'-diphosphate
      `A3P`_   adenosine-3',5'-diphosphate

      `CTN`_   cytidine
      `C2P`_   cytidine-2'-monophosphate
      `C3P`_   cytidine-3'-monophosphate
      `C5P`_   cytidine-5'-monophosphate
      `CDP`_   cytidine-5'-diphosphate
      `CTP`_   cytidine-5'-triphosphate

      `GMP`_   guanosine
      `5GP`_   guanosine-5'-monophosphate
      `GDP`_   guanosine-5'-diphosphate
      `GTP`_   guanosine-5'-triphosphate

      `THM`_   thymidine
      `TMP`_   thymidine-5'-monophosphate
      `TPP`_   thymidine-5'-diphosphate
      `TTP`_   thymidine-5'-triphosphate

      `URI`_   uridine (uracil plus ribose)
      `UMP`_   2'-deoxyuridine 5'-monophosphate
      `UDP`_   uridine 5'-diphosphate
      `UTP`_   uridine 5'-triphosphate
      =======  ================================


   at
      same as selection ``'resname ADE A THY T'``

   cg
      same as selection ``'resname CYT C GUN G'``

   purine
      same as selection ``'resname ADE A GUN G'``

   pyrimidine
      same as selection ``'resname CYT C THY T URA U'``

"""

for resi in DEFAULTS['nucleobase']:
    __doc__ += PDBLIGSUM.format(resi)

for resi in DEFAULTS['nucleotide']:
    __doc__ += PDBLIGSUM.format(resi)

for resi in DEFAULTS['nucleoside']:
    __doc__ += PDBLIGSUM.format(resi)

__doc__ += """

Heteros
-------------------------------------------------------------------------------

.. glossary::

   hetero
      indicates anything other than a :term:`protein` or a
      :term:`nucleic` residue, i.e. ``'not (protein or nucleic)'``.


   hetatm
      is available when atomic data is parsed from a PDB or similar
      format file and indicates atoms that are marked ``'HETATM'`` in the file.


   water
      indices `HOH`_ and `DOD`_ recognized by *PDB* and also WAT, TIP3, H2O,
      OH2, TIP, TIP2, TIP4 and SPC recognized by molecular dynamics (MD) force
      fields.

      .. _HOH: http://www.pdb.org/pdb/ligand/ligandsummary.do?hetId=HOH

      .. _DOD: http://www.pdb.org/pdb/ligand/ligandsummary.do?hetId=DOD

      Previously used water types HH0, OHH, and SOL conflict with other
      compounds in the *PDB*, so are removed from the definition of this flag
      except SOL (restored) as compound SOL (L-sorbose) is only used 3 times.


   ion
      indicates the following ions most of which are recognized by the *PDB*
      and others by MD force fields.

      =======  ==================  =====  ========  ==========
      \                            *PDB*  *Source*  *Conflict*
      `AL`_    aluminum            Yes
      `BA`_    barium              Yes
      `CA`_    calcium             Yes
      `CD`_    cadmium             Yes
      `CL`_    chloride            Yes
      `CO`_    cobalt (ii)         Yes
      `CS`_    cesium              Yes
      `CU`_    copper (ii)         Yes
      `CU1`_   copper (i)          Yes
      `CUA`_   dinuclear copper    Yes
      `HG`_    mercury (ii)        Yes
      `IN`_    indium (iii)        Yes
      `IOD`_   iodide              Yes
      `K`_     potassium           Yes
      `MG`_    magnesium           Yes
      `MN3`_   manganese (iii)     Yes
      `MN`_    manganese (ii)      Yes    
      `NA`_    sodium              Yes
      `PB`_    lead (ii)           Yes
      `PT`_    platinum (ii)       Yes
      `RB`_    rubidium            Yes
      `TB`_    terbium (iii)       Yes
      `TL`_    thallium (i)        Yes
      `WO4`_   thungstate (vi)     Yes
      `YB`_    ytterbium (iii)     Yes
      `ZN`_    zinc                Yes
      CAL      calcium             No     CHARMM    Yes
      CES      cesium              No     CHARMM    Yes
      CLA      chloride            No     CHARMM    Yes
      POT      potassium           No     CHARMM    Yes
      SOD      sodium              No     CHARMM    Yes
      ZN2      zinc                No     CHARMM    No
      CU2P     copper (ii)         No     CHARMM    No
      CU2      copper (ii)         No     CHARMM    No
      =======  ==================  =====  ========  ==========

      Ion identifiers that are obsoleted by *PDB* (MO3, MO4, MO5, MO6, NAW,
      OC7, and ZN1) are removed from this definition. 

      CU2 comes from CU2P if parsing PDB files without long_resname
      or writing them again (always trims to 3-character resnames).


   lipid
      {lipid}


   sugar
      {sugar}


   heme
      {heme}

   pdbter
      is available when atomic data is parsed from a PDB format file and
      indicates atoms that were followed by ``'TER'`` record.

   selpdbter
      is available when atomic data is parsed from a PDB format file and
      then a selection is made and indicates selected atoms that should
      be followed by ``'TER'`` record.

""".format(

lipid = wrapText('indicates ' +
                 joinLinks(DEFAULTS['lipid'], last=', and ', sort=True) +
                 ' from *PDB*, and also ' + ', '.join(DEFAULTS['lipid_other']) +
                 ' from *CHARMM* force field.', subsequent_indent=' '*6),

sugar = wrapText('indicates ' +
                 joinLinks(DEFAULTS['sugar'], last=', and ', sort=True) +
                 ' from *PDB*, and also AGLC from *CHARMM*.',
                 subsequent_indent=' '*6),

heme = wrapText('indicates ' +
                joinLinks(DEFAULTS['heme'], last=', and ', sort=True) +
                ' from *PDB*, and also HEMO and HEMR from CHARMM.',
                subsequent_indent=' '*6)
)


for resi in DEFAULTS['ion']:
    __doc__ += PDBLIGSUM.format(resi)

for resi in DEFAULTS['lipid']:
    __doc__ += PDBLIGSUM.format(resi)

for resi in DEFAULTS['sugar']:
    __doc__ += PDBLIGSUM.format(resi)

for resi in DEFAULTS['heme']:
    __doc__ += PDBLIGSUM.format(resi)


__doc__ += """

Elements
-------------------------------------------------------------------------------

Following elements found in proteins are recognized by applying regular
expressions to atom names:

.. glossary::

   carbon
      carbon atoms, same as ``'name "C.*" and not ion'``

   nitrogen
      nitrogen atoms, same as ``'name "N.*" and not ion'``

   oxygen
      oxygen atoms, same as ``'name "O.*" and not ion'``

   sulfur
      sulfur atoms, same as ``'name "S.*" and not ion'``

   hydrogen
      hydrogen atoms, same as ``'name "[1-9]?H.*" and not ion'``

   noh
   heavy
      non hydrogen atoms, same as ``'not hydrogen``


``'not ion'`` is appended to above definitions to avoid conflicts with
:term:`ion` atoms.

Structure
-------------------------------------------------------------------------------

Following secondary structure flags are defined but before they can be used,
secondary structure assignments must be made.

.. glossary::

   extended
      extended conformation, same as ``'secondary E'``

   helix
      α-helix conformation, same as ``'secondary H'``

   helix310
      3_10-helix conformation, same as ``'secondary G'``

   helixpi
      π-helix conformation, same as ``'secondary I'``

   turn
      hydrogen bonded turn conformation, same as ``'secondary T'``

   bridge
      isolated beta-bridge conformation, same as ``'secondary B'``

   bend
      bend conformation, same as ``'secondary S'``

   coil
      not in one of above conformations, same as ``'secondary C'``

Others
-------------------------------------------------------------------------------

.. glossary::

   all
      indicates all atoms, returns a new view of the instance

   none
      indicates no atoms, returns **None**

   dummy
      indicates dummy atoms in an :class:`.AtomMap`

   mapped
      indicates mapped atoms in an :class:`.AtomMap`

Functions
===============================================================================

The following functions can be used to customize flag definitions:

    * :func:`.flagDefinition`
    * :func:`.addNonstdAminoacid`
    * :func:`.delNonstdAminoacid`
    * :func:`.listNonstdAAProps`

"""

# join split lists
for key in ['ion', 'lipid', 'sugar', 'heme']:
    DEFAULTS[key].update(DEFAULTS.pop(key + '_other'))

DEFINITIONS = None
AMINOACIDS = None
BACKBONE = None

MODMAP = {}
with openData('mod_res_map.dat') as f:
    for line in f:
        try:
            mod, aa = line.strip().split(' ')
            MODMAP[mod] = aa
        except:
            continue

NAMAP = {'ADE': 'a', 'THY': 't', 'CYT': 'c',
         'GUA': 'g', 'URA': 'u'}

# add modified bases to NAMAP
MODNAMAP = {}
for mod, aa in MODMAP.items():
    if aa in NAMAP:
        MODNAMAP[mod] = NAMAP[aa]
NAMAP.update(MODNAMAP)


def updateDefinitions():
    """Update definitions and set some global variables.  This function must be
    called at the end of the module."""

    global DEFINITIONS, AMINOACIDS, BACKBONE, TIMESTAMP
    DEFINITIONS = {}
    user = SETTINGS.get('flags_definitions', {})

    # nucleics
    nucleic = set()
    for key in ['nucleobase', 'nucleoside', 'nucleotide']:
        aset = set(user.get(key, DEFAULTS[key]))
        nucleic.update(aset)
        DEFINITIONS[key] = aset
    for key in NAMAP:
        nucleic.update(set(NAMAP.keys()))
    DEFINITIONS['nucleic'] = nucleic

    # heteros
    for key in ['water', 'lipid', 'ion', 'sugar', 'heme',
                'at', 'cg', 'purine', 'pyrimidine']:
        DEFINITIONS[key] = set(user.get(key, DEFAULTS[key]))

    DEFINITIONS['backbone'] = DEFINITIONS['bb'] = set(user.get(key,
                                                      DEFAULTS['bb']))
    DEFINITIONS['backbonefull'] = DEFINITIONS['bbfull'] = set(user.get(key,
                                                           DEFAULTS['bbfull']))

    # element regex
    for key in ['hydrogen', 'carbon', 'nitrogen', 'oxygen', 'sulfur']:
        DEFINITIONS[key] = recompile(user.get(key, DEFAULTS[key]))

    try:
        nonstd = SETTINGS[NONSTANDARD_KEY]

    except KeyError:
        nonstd = NONSTANDARD
        DEFINITIONS.update(CATEGORIZED)
    else:

        for cat in CATEGORIES:
            for key in CATEGORIES[cat]:
                DEFINITIONS[key] = set(DEFAULTS[key])

        DEFINITIONS['charged'] = set(DEFINITIONS['acidic'])
        DEFINITIONS['charged'].update(DEFINITIONS['basic'])

        for resi, props in nonstd.items():
            for prop in props:
                DEFINITIONS[prop].add(resi)

    DEFINITIONS['stdaa'] = DEFAULTS['stdaa']
    DEFINITIONS['nonstdaa'] = set(nonstd)
    AMINOACIDS = set(DEFINITIONS['stdaa'])
    AMINOACIDS.update(DEFINITIONS['nonstdaa'])
    DEFINITIONS['protein'] = DEFINITIONS['aminoacid'] = AMINOACIDS

    BACKBONE = DEFINITIONS['bb']

    global TIMESTAMP
    TIMESTAMP = SETTINGS.get(TIMESTAMP_KEY, 0)


#==============================================================================
# PLANTERS
#==============================================================================

# flag planters are functions that take an AtomGroup instance
# and set flags for a given label using the atomic data stored in the AtomGroup
# a planter can also set the indices of subset atoms for calculated flag
# planters return the flags after calculation is over

# protein
#==============================================================================

def setCalpha(ag, label):

    flags = ag._getNames() == 'CA'
    indices = flags.nonzero()[0]
    if len(indices):
        torf = array([rn in AMINOACIDS for rn in ag._getResnames()[indices]])
        flags[indices] = torf
        ag._setFlags(label, flags)
        ag._setSubset(label, indices[torf])
    else:
        ag._setFlags(label, flags)
        ag._setSubset(label, array([]))
    return flags

addPlanter(setCalpha, 'ca', 'calpha')


def setProtein(ag, label):

    calpha = ag._getSubset('calpha')
    if len(calpha):
        torf = zeros(ag.numResidues(), bool)
        resindices = ag._getResindices()
        torf[resindices[calpha]] = True
        flags = torf[resindices]
    else:
        flags = zeros(ag.numAtoms(), bool)
        
    water = ag._getSubset("water")
    if len(water):
        flags[water] = False
        
    ions = ag._getSubset("ion")
    if len(ions):
        flags[ions] = False
    
    ag._setFlags('protein', flags)
    return flags

addPlanter(setProtein, 'protein', 'aminoacid')

# subsets
#==============================================================================


def setBackbone(ag, label):

    protein = ag._getSubset('protein')
    if len(protein):
        bb = DEFINITIONS[label]
        names = ag._getNames()[protein]
        flags = zeros(ag.numAtoms(), bool)
        flags[protein] = [nm in bb for nm in names]
    else:
        flags = zeros(ag.numAtoms(), bool)
    ag._setFlags(label, flags)
    return flags

addPlanter(setBackbone, 'bb', 'backbone', editor=changeBackbone)
addPlanter(setBackbone, 'bbfull', 'backbonefull', editor=changeBackbone)


def setSidechain(ag, label):

    protein = ag._getSubset('protein')
    if len(protein):
        flags = zeros(ag.numAtoms(), bool)
        flags[protein] = True
        flags[ag._getSubset('bbfull')] = False
    else:
        flags = zeros(ag.numAtoms(), bool)
    ag._setFlags(label, flags)
    return flags

addPlanter(setSidechain, 'sc', 'sidechain')


def setCategories(ag, label):

    calpha = ag._getSubset('ca')
    if len(calpha):
        resnames = DEFINITIONS[label]
        residx = ag._getResindices()
        torf = zeros(ag.numResidues(), bool)
        torf[residx[calpha]] = [rn in resnames
                                for rn in ag._getResnames()[calpha]]
        flags = torf[residx]
    else:
        flags = zeros(ag.numAtoms(), bool)
    ag._setFlags(label, flags)
    return flags

addPlanter(setCategories, 'stdaa', 'nonstdaa', *list(CATEGORIZED.keys()),
           aliases=False)


def setAll(ag, label):

    flags = ones(ag.numAtoms(), bool)
    ag._setFlags('all', flags)
    return flags

addPlanter(setAll, 'all')


def setNone(ag, label):

    flags = zeros(ag.numAtoms(), bool)
    ag._setFlags('none', flags)
    return flags

addPlanter(setNone, 'none')

# hetero, nucleic, water, etc.
#==============================================================================


def setResiflag(ag, label):

    resnames = DEFINITIONS[label]
    flags = array([rn in resnames for rn in ag._getResnames()], bool)
    ag._setFlags(label, flags)
    return flags

addPlanter(setResiflag, 'nucleobase', 'nucleoside', 'nucleotide',
           'water', 'ion', 'lipid', 'sugar', 'heme', 'at', 'cg', 'purine',
           'pyrimidine', aliases=False, editor=changeResnames)
addPlanter(setResiflag, 'nucleic')


def setHetero(ag, label):

    flags = ones(ag.numAtoms(), bool)
    flags[ag._getSubset('protein')] = False
    flags[ag._getSubset('nucleic')] = False
    ag._setFlags('hetero', flags)
    return flags

addPlanter(setHetero, 'hetero',)


# element
#==============================================================================

def setElement(ag, label):

    match = DEFINITIONS[label].match
    flags = array([match(nm) is not None for nm in ag._getNames()])
    flags[ag._getSubset('ion')] = False
    ag._setFlags(label, flags)
    return flags

addPlanter(setElement, 'hydrogen', 'carbon', 'nitrogen', 'oxygen', 'sulfur',
           aliases=False, editor=changeNameRegex)


def setNoh(ag, label):

    flags = ones(ag.numAtoms(), bool)
    flags[ag._getSubset('hydrogen')] = False
    ag._setFlags(label, flags)
    return flags

addPlanter(setNoh, 'noh', 'heavy')


# secondary
#==============================================================================

SECONDARY = {'extended': 'E', 'helix': 'H', 'helix310': 'G', 'helixpi': 'I',
             'turn': 'T', 'bridge': 'B', 'bend': 'S', 'coil': 'C'}


def setSecondary(ag, label):

    secstrs = ag._getSecstrs()
    if secstrs is None:
        raise ValueError('no secondary structure assignments')
    flags = secstrs == SECONDARY[label]
    ag._setFlags(label, flags)
    return flags

addPlanter(setSecondary, *list(SECONDARY.keys()), aliases=False,
           fields=['secondary'])


# secondary
#==============================================================================

def setDummy(ag, label):

    if label == 'dummy':
        flags = zeros(len(ag), bool)
    else:
        flags = ones(len(ag), bool)
    ag._setFlags(label, flags)
    return flags

addPlanter(setDummy, 'dummy', 'mapped', aliases=False, fields=[])


#==============================================================================
# User functions
#==============================================================================

def flagDefinition(*arg, **kwarg):
    """Learn, change, or reset :ref:`flags` definitions.

    **Learn a definition**

    Calling this function with no arguments will return list of flag names
    whose definitions you can learn:

    .. ipython:: python

       flagDefinition()

    Passing a flag name will return its definition:

    .. ipython:: python

       flagDefinition('backbone')
       flagDefinition('hydrogen')

    **Change a definition**

    Calling the function with ``editable=True`` argument will return flag
    names those definitions that can be edited:

    .. ipython:: python

       flagDefinition(editable=True)

    Pass an editable flag name with its new definition:

    .. ipython:: python

       flagDefinition(nitrogen='N.*')
       flagDefinition(backbone=['CA', 'C', 'O', 'N'])
       flagDefinition(nucleobase=['ADE', 'CYT', 'GUN', 'THY', 'URA'])

    Note that the type of the new definition must be the same as the type
    of the old definition.  Flags with editable definitions are:
    {editable}

    **Reset definitions**

    Pass *reset* keyword as follows to restore all default definitions of
    editable flags and also non-standard amino acids.

    .. ipython:: python

       flagDefinition(reset='all')

    Or, pass a specific editable flag label to restore its definition:

    .. ipython:: python

       flagDefinition(reset='nitrogen')"""

    if arg and kwarg:
        raise ValueError('only a single argument or a single keyword '
                         'argument is accepted at a time')

    elif arg:
        if len(arg) > 1:
            raise ValueError('specify only one flag label')
        arg = arg[0]
        try:
            definition = DEFINITIONS[arg]
        except KeyError:
            raise ValueError('{0} is not a valid flag label'
                             .format(repr(arg)))
        else:
            try:
                return definition.pattern
            except AttributeError:
                definition = list(definition)
                definition.sort()
                return definition

    elif kwarg:
        if len(kwarg) > 1:
            raise ValueError('specify only one keyword argument')
        key, val = list(kwarg.items())[0]
        if key == 'editable' and val:
            alist = list(EDITORS)
            alist.sort()
            return alist
        if key == 'reset' and val:
            resetDefinitions(val)
            return
        try:
            editor = EDITORS[key]
        except KeyError:
            raise ValueError('{0} is not an editable flag or a valid '
                             'keyword'.format(repr(key)))
        else:
            type_ = type(flagDefinition(key))
            if type(val) != type_:
                raise TypeError('expected {0} as type of new definition, '
                                'found {1}'.format(repr(type_.__name__),
                                repr(type(val).__name__)))
            editor(key, val)
    else:
        alist = list(DEFINITIONS)
        alist.sort()
        return alist

flagDefinition.__doc__ = flagDefinition.__doc__.format(
    editable=wrapText(joinTerms(flagDefinition(editable=True), last=', and '),
                      subsequent_indent=' '*4))


def addNonstdAminoacid(resname, *properties):
    """Add non-standard amino acid *resname* with *properties* selected from:

      * {props}

    .. ipython:: python

       addNonstdAminoacid('PTR', 'acidic', 'aromatic', 'cyclic', 'large',
       'polar', 'surface')

    Default set of non-standard amino acids can be restored as follows:

    .. ipython:: python

       flagDefinition(reset='nonstdaa')"""

    resname = str(resname)
    if len(resname) > 4:
        LOGGER.warn('Residue name {0} is unusually long.'
                    .format(repr(resname)))
    propset = set(properties)
    for cat, val in CATEGORIES.items():
        intersection = val.intersection(propset)
        if intersection:
            if len(intersection) > 1:
                raise ValueError('amino acid properties {0} cannot be '
                                 'present together'
                                 .format(', '.join([repr(prp)
                                                    for prp in intersection])))
            for prop in intersection:
                propset.remove(prop)
    if propset:
        raise ValueError('amino acid property {0} is not valid'
                         .format(repr(propset.pop())))

    nonstd = SETTINGS.get(NONSTANDARD_KEY, NONSTANDARD)
    nonstd[resname] = set(properties)
    updateNonstandard(nonstd)

INDEPENDENT.sort()
addNonstdAminoacid.__doc__ = addNonstdAminoacid.__doc__.format(
    props='\n      * '.join([
        '*{0}*: {1}'.format(cat, joinTerms(terms, last=', or ', sort=True))
        for cat, terms in CATEGORIES.items()
    ])
)


def delNonstdAminoacid(resname):
    """Delete non-standard amino acid *resname*.

    .. ipython:: python

       delNonstdAminoacid('PTR')
       flagDefinition('nonstdaa')

    Default set of non-standard amino acids can be restored as follows:

    .. ipython:: python

       flagDefinition(reset='nonstdaa')"""

    nonstd = SETTINGS.get(NONSTANDARD_KEY, NONSTANDARD)
    try:
        nonstd.pop(resname)
    except KeyError:
        raise ValueError('{0} is not a non-standard residue name'
                         .format(repr(resname)))
    else:
        updateNonstandard(nonstd)


def getNonstdProperties(resname):
    """Deprecated for removal in v1.4, use :func:`listNonstdAAProps` instead.
    """

    from prody import deprecate
    deprecate('getNonstdProperties', 'listNonstdAAProps')

    return listNonstdAAProps(resname)


def listNonstdAAProps(resname):
    """Returns properties of non-standard amino acid *resname*.

    .. ipython:: python

       listNonstdAAProps('PTR')"""

    try:
        alist = list(SETTINGS.get(NONSTANDARD_KEY, NONSTANDARD)[resname])
    except KeyError:
        raise ValueError('{0} is not a non-standard residue name'
                         .format(repr(resname)))
    else:
        alist.sort()
        return alist


updateDefinitions()



from numpy import array, concatenate, unique
from numpy.testing import assert_array_equal, assert_equal

from prody.tests import unittest
from prody.tests.datafiles import parseDatafile, pathDatafile

from prody.measure import Contacts, findNeighbors, iterNeighbors
from prody.measure import buildDistMatrix, calcDistance


UBI = parseDatafile('1ubi')
UBI_XYZ = UBI._getCoords()
UBI_MIN = UBI_XYZ.min(0)
UBI_MAX = UBI_XYZ.max(0)
UBI_UC = UBI_MAX - UBI_MIN
N0P = [-1, 0, 1]
UBI_EDGES = UBI_MIN + UBI_UC * array([(i, j, k) for i in N0P
                                                for j in N0P for k in N0P])
UBI_RADIUS = 20

UBI_CONTACTS = Contacts(UBI_XYZ)
UBI_CONTACTS_PBC = Contacts(UBI_XYZ, UBI_UC)

UBI_ADDH = parseDatafile('1ubi_addH')


class TestContacts(unittest.TestCase):

    """Test functions that calculate dihedral angles."""

    def testPBCvsNONE(self):

        wout_pbc = unique(concatenate([UBI_CONTACTS(UBI_RADIUS, UBI_EDGES)]))
        with_pbc = UBI_CONTACTS_PBC(UBI_RADIUS, UBI_MIN)
        assert_array_equal(wout_pbc, with_pbc)

    def testMINvsMAX(self):

        assert_array_equal(UBI_CONTACTS_PBC(UBI_RADIUS, UBI_MIN),
                           UBI_CONTACTS_PBC(UBI_RADIUS, UBI_MAX))


UCA = UBI.ca.copy()[::2].copy()
UCA_XYZ = UCA._getCoords()
UCA_MIN = UCA_XYZ.min(0)
UCA_MAX = UCA_XYZ.max(0)
UCA_UC = UCA_MAX - UCA_MIN
N0P = [-1, 0, 1]
UCA_EDGES = UCA_MIN + UCA_UC * array([(i, j, k) for i in N0P
                                                for j in N0P for k in N0P])
UCA_RADIUS = 20

UCA_CONTACTS = Contacts(UCA_XYZ)
UCA_CONTACTS_PBC = Contacts(UCA_XYZ, UCA_UC)

class TestNeighbors(unittest.TestCase):

    def testPBC(self):

        neighbors = findNeighbors(UCA_XYZ, UCA_RADIUS, unitcell=UCA_UC)
        n_neighbors = (buildDistMatrix(UCA_XYZ, unitcell=UCA_UC,
                                       format='arr') <= UCA_RADIUS).sum()
        assert_equal(len(neighbors), n_neighbors)

    def testNoPBC(self):

        neighbors = findNeighbors(UCA_XYZ, UCA_RADIUS)
        n_neighbors = (buildDistMatrix(UCA_XYZ,
                                       format='arr') <= UCA_RADIUS).sum()
        assert_equal(len(neighbors), n_neighbors)

    def testCoordArgumentSwitching(self):

        dist = 12.
        neighbors1 = findNeighbors(UCA_XYZ, dist, UCA_XYZ[1])
        neighbors2 = findNeighbors(UCA_XYZ[1], dist, UCA_XYZ)
        n_neighbors = (calcDistance(UCA_XYZ, UCA_XYZ[1]) <= dist).sum()
        assert_equal(len(neighbors1), n_neighbors)
        neighbors1.sort()
        neighbors2.sort()
        assert_array_equal(array(neighbors1)[:,-1],
                           array(neighbors2)[:,-1])

    def testAtomicArgumentSwitching(self):

        dist = 12.
        neighbors1 = [(a.getIndex(), d)
                      for a, b, d in iterNeighbors(UCA, dist, UCA[1])]
        neighbors2 = [(b.getIndex(), d)
                      for a, b, d in iterNeighbors(UCA[1], dist, UCA)]
        n_neighbors = (calcDistance(UCA, UCA[1]) <= dist).sum()
        assert_equal(len(neighbors1), n_neighbors)
        neighbors1.sort()
        neighbors2.sort()
        self.assertEqual(neighbors1, neighbors2)

    def testSeqsep(self):

        neighbors = findNeighbors(UCA, UCA_RADIUS, seqsep=3)
        n_neighbors = (buildDistMatrix(UCA, seqsep=3, 
                                       format='arr') <= UCA_RADIUS).sum()
        assert_equal(len(neighbors), n_neighbors)

    def testPBCCoordArgumentSwitching(self):

        dist = 12.
        neighbors1 = findNeighbors(UCA_XYZ, dist, UCA_XYZ[1], unitcell=UCA_UC)
        neighbors2 = findNeighbors(UCA_XYZ[1], dist, UCA_XYZ, unitcell=UCA_UC)
        n_neighbors = (calcDistance(UCA_XYZ, UCA_XYZ[1], unitcell=UCA_UC) <=
                       dist).sum()
        assert_equal(len(neighbors1), n_neighbors)
        neighbors1.sort()
        neighbors2.sort()
        assert_array_equal(array(neighbors1)[:,-1],
                           array(neighbors2)[:,-1])

    def testPBCAtomicArgumentSwitching(self):

        dist = 12.
        neighbors1 = [(a.getIndex(), d) for a, b, d in
                            iterNeighbors(UCA, dist, UCA[1], unitcell=UCA_UC)]
        neighbors2 = [(b.getIndex(), d) for a, b, d in
                            iterNeighbors(UCA[1], dist, UCA, unitcell=UCA_UC)]
        n_neighbors = (calcDistance(UCA, UCA[1], unitcell=UCA_UC) <=
                       dist).sum()
        assert_equal(len(neighbors1), n_neighbors)
        neighbors1.sort()
        neighbors2.sort()
        self.assertEqual(neighbors1, neighbors2)



# -*- coding: utf-8 -*-
"""This module defines atom pointer base class."""

from numbers import Integral
from numpy import all, array, concatenate, ones, unique, any

from .atomic import Atomic
from .bond import Bond
from .angle import Angle
from .dihedral import Dihedral
from .improper import Improper
from .donor import Donor
from .acceptor import Acceptor
from .crossterm import Crossterm
from .nbexclusion import NBExclusion

from prody import LOGGER

__all__ = ['AtomPointer']


class AtomPointer(Atomic):

    """A base for classes pointing to atoms in :class:`.AtomGroup` instances.
    Derived classes are:

      * :class:`.Atom`
      * :class:`.AtomSubset`
      * :class:`.AtomMap`"""

    __slots__ = ['_ag', '_acsi']

    def __init__(self, ag, acsi):

        if not isinstance(ag, AtomGroup):
            raise TypeError('ag must be an AtomGroup instance, not {0}'
                            .format(type(ag)))
        self._ag = ag
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = int(acsi)

    def __contains__(self, item):

        try:
            ag = item.getAtomGroup()
        except AttributeError:
            return False
        else:
            return (self._ag == ag and len(item) <= len(self) and
                    set(item._getIndices()).issubset(set(self._getIndices())))

    def __eq__(self, other):

        try:
            ag = other.getAtomGroup()
        except AttributeError:
            return False
        else:
            return (self._ag == ag and other.numAtoms() == self.numAtoms() and
                    all(self._getIndices() == other._getIndices()))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __invert__(self):

        torf = ones(self._ag.numAtoms(), bool)
        torf[self._indices] = False
        return Selection(self._ag, torf.nonzero()[0],
                         'not ({0})'.format(self.getSelstr()),
                         self.getACSIndex(), unique=True)

    def __or__(self, other):

        if self is other:
            return self

        try:
            ag = other.getAtomGroup()
        except AttributeError:
            raise TypeError('other must be an AtomPointer')

        if self._ag != ag:
            raise ValueError('both selections must be from the same AtomGroup')

        acsi = self.getACSIndex()
        if acsi != other.getACSIndex():
            LOGGER.warn('Active coordinate set indices do not match, it will '
                        'be set to zero.')
            acsi = 0

        indices = unique(concatenate((self._getIndices(),
                                      other._getIndices())))
        if indices[-1] == atommap.DUMMY:
            indices = indices[:-1]
        return Selection(self._ag, indices, '({0}) or ({1})'
                         .format(self.getSelstr(), other.getSelstr()),
                         acsi, unique=True)

    def __and__(self, other):

        if self is other:
            return self

        try:
            ag = other.getAtomGroup()
        except AttributeError:
            raise TypeError('other must be an AtomPointer')

        if self._ag != ag:
            raise ValueError('both selections must be from the same AtomGroup')

        acsi = self.getACSIndex()
        if acsi != other.getACSIndex():
            LOGGER.warning('active coordinate set indices do not match, '
                           'so it will be set to zero in the union.')
            acsi = 0

        acsi = self.getACSIndex()
        if acsi != other.getACSIndex():
            LOGGER.warn('Active coordinate set indices do not match, it will '
                        'be set to zero.')
            acsi = 0

        indices = set(self._getIndices())

        indices = indices.intersection(other.getIndices())
        if indices:
            indices = unique(indices)
            if indices[-1] == atommap.DUMMY:
                indices = indices[:-1]
            return Selection(self._ag, indices, '({0}) and ({1})'
                             .format(self.getSelstr(), other.getSelstr()),
                             acsi)

    def __add__(self, other):
        """Returnss an :class:`.AtomMap` instance. Order of pointed atoms are
        preserved."""

        try:
            ag = other.getAtomGroup()
        except AttributeError:
            raise TypeError('unsupported operand type(s) for +: {0} and '
                            '{1}'.format(repr(type(self).__name__),
                                         repr(type(other).__name__)))

        if ag != self._ag:
            raise ValueError('AtomPointer instances must point to the same '
                             'AtomGroup instance')
        acsi = self.getACSIndex()
        if acsi != other.getACSIndex():
            LOGGER.warning('Active coordset indices of atoms are not the same.'
                           ' Result will have ACSI {0}.'.format(acsi))

        title = '({0}) + ({1})'.format(str(self), str(other))
        indices = concatenate([self._getIndices(), other._getIndices()])

        dummies = 0
        try:
            dummies += self.numDummies()
        except AttributeError:
            pass
        try:
            dummies += other.numDummies()
        except AttributeError:
            pass

        return AtomMap(ag, indices, acsi, title=title, intarrays=True,
                       dummies=dummies)

    def _getTimeStamp(self, index=None):

        if index is None:
            index = self.getACSIndex()
        return self._ag._getTimeStamp(index)

    def _getKDTree(self):
        """Returns KDTree for the active coordinate set from the atom group."""

        return self._ag._getKDTree(self.getACSIndex())

    def getAtomGroup(self):
        """Returns associated atom group."""

        return self._ag

    def numCoordsets(self):
        """Returns number of coordinate sets."""

        return self._ag._n_csets

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set coordinates at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index

    def getACSLabel(self):
        """Returns active coordinate set label."""

        if self._ag._n_csets:
            return self._ag._cslabels[self.getACSIndex()]

    def getCSLabels(self):
        """Returns coordinate set labels."""

        return self._ag.getCSLabels()

    def isDataLabel(self, label):
        """Returns **True** if data associated with *label* is present."""

        return self._ag.isDataLabel(label)

    def getDataLabels(self, which=None):
        """Returns data labels.  For ``which='user'``, return only labels of
        user provided data."""

        return self._ag.getDataLabels(which)

    def getDataType(self, label):
        """Returns type of the data (i.e. ``data.dtype``) associated with
        *label*, or **None** label is not used."""

        return self._ag.getDataType(label)

    def getFlagLabels(self, which=None):
        """Returns flag labels.  For ``which='user'``,  return labels of user
        or parser (e.g. :term:`hetatm`) provided flags, for ``which='all'``
        return all possible :ref:`flags` labels in addition to those present
        in the instance."""

        return self._ag.getFlagLabels(which)

    def isFlagLabel(self, label):
        """Returns **True** if flags associated with *label* are present."""

        return self._ag.isFlagLabel(label)

    def _getFlags(self, label):
        """Returns atom flags."""

        flags = self._ag._getFlags(label)
        if flags is not None:
            return flags[self._getIndices()]

    def _getSubset(self, label):

        subset = array(list(set(self._ag._getSubset(label))
                            .intersection(set(self._getIndices()))), int)
        subset.sort()
        return subset
    
    def getAnisous(self):
        """Returns a copy of anisotropic temperature factors from the active coordinate set."""

        if self._ag._anisous is not None:
            # Since this is not slicing, a view is not returned
            return self._ag._anisous[self.getACSIndex(), self._indices]

    _getAnisous = getAnisous

    def getBonds(self):
        """Returns bonds.  Use :meth:`setBonds` or
        :meth:`inferBonds` from parent AtomGroup for setting bonds."""

        if self._ag._bonds is not None:
            iset = set(self._getIndices())
            acsi = self._acsi
            return array([Bond(self, bond, acsi) for bond in self._ag._bonds
                          if bond[0] in iset and bond[1] in iset])
        return None

    def numBonds(self):
        """Returns number of bonds.  Use :meth:`setBonds` or
        :meth:`inferBonds` from parent AtomGroup for setting bonds."""
        return len(self.getBonds())

    def _iterBonds(self):
        """Yield pairs of indices for bonded atoms that are within the pointer.
        Use :meth:`setBonds` from parent AtomGroup for setting bonds."""

        if self._ag._bonds is None:
            LOGGER.warning('bonds are not set, use `setBonds` or `inferBonds`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 2 >= len(self):
            for a, b in self._ag._iterBonds():
                if a in iset and b in iset:
                    yield a, b
        else:
            if any(self._ag._bmap):
                for a, bmap in zip(indices, self._ag._bmap[indices]):
                    for b in bmap:
                        if b > -1 and b in iset:
                            yield a, b
                    iset.remove(a)

    def iterBonds(self):
        """Yield bonds formed by the atom.  Use :meth:`setBonds` or 
        :meth:`inferBonds` for setting bonds."""

        ag = self._ag
        acsi = self.getACSIndex()
        for bond in self._iterBonds():
            yield Bond(ag, bond, acsi)

    def _iterAngles(self):
        """Yield triplets of indices for angled atoms that are within the pointer.
        Use :meth:`setAngles` for setting angles."""

        if self._ag._angles is None:
            LOGGER.warning('angles are not set, use `AtomGroup.setAngles`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 3 >= len(self):
            for a, b, c in self._ag._iterAngles():
                if a in iset and b in iset and c in iset:
                    yield a, b, c
        else:
            if any(self._ag._angmap):
                for a, amap in zip(indices, self._ag._angmap[indices]):
                    for b, c in amap:
                        if b > -1 and b in iset and c > -1 and c in iset:
                            yield a, b, c
                    iset.remove(a)

    def iterAngles(self):
        """Yield angles formed by the atom.  Use :meth:`setAngles` for setting
        angles."""

        ag = self._ag
        acsi = self.getACSIndex()
        for angle in self._iterAngles():
            yield Angle(ag, angle, acsi)

    def _iterDihedrals(self):
        """Yield quadruples of indices for dihedraled atoms that are within the pointer.
        Use :meth:`setDihedrals` for setting dihedrals."""

        if self._ag._dihedrals is None:
            LOGGER.warning('dihedrals are not set, use `AtomGroup.setDihedrals`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 4 >= len(self):
            for a, b, c, d in self._ag._iterDihedrals():
                if a in iset and b in iset and c in iset and d in iset:
                    yield a, b, c, d
        else:
            if any(self._ag._dmap):
                for a, dmap in zip(indices, self._ag._dmap[indices]):
                    for b, c, d in dmap:
                        if b > -1 and b in iset and c > -1 and c in iset \
                        and d > -1 and d in iset:
                            yield a, b, c, d
                    iset.remove(a)

    def iterDihedrals(self):
        """Yield dihedrals formed by the atom.  Use :meth:`setDihedrals` for setting
        dihedrals."""

        ag = self._ag
        acsi = self.getACSIndex()
        for dihedral in self._iterDihedrals():
            yield Dihedral(ag, dihedral, acsi) 

    def _iterImpropers(self):
        """Yield quadruplet of indices for impropered atoms that are within the pointer.
        Use :meth:`setImpropers` for setting impropers."""

        if self._ag._impropers is None:
            LOGGER.warning('impropers are not set, use `AtomGroup.setImpropers`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 4 >= len(self):
            for a, b, c, d in self._ag._iterImpropers():
                if a in iset and b in iset and c in iset and d in iset:
                    yield a, b, c, d
        else:
            if any(self._ag._imap):
                for a, imap in zip(indices, self._ag._imap[indices]):
                    for b, c, d in imap:
                        if b > -1 and b in iset and c > -1 and c in iset \
                        and d > -1 and d in iset:
                            yield a, b, c, d
                    iset.remove(a)

    def iterImpropers(self):
        """Yield impropers formed by the atom.  Use :meth:`setImpropers` for setting
        impropers."""

        ag = self._ag
        acsi = self.getACSIndex()
        for improper in self._iterImpropers():
            yield Improper(ag, improper, acsi) 

    def _iterCrossterms(self):
        """Yield quadruplet of indices for crosstermed atoms that are within the pointer.
        Use :meth:`setCrossterms` for setting crossterms."""

        if self._ag._crossterms is None:
            LOGGER.warning('crossterms are not set, use `AtomGroup.setCrossterms`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 4 >= len(self):
            for a, b, c, d in self._ag._iterCrossterms():
                if a in iset and b in iset and c in iset and d in iset:
                    yield a, b, c, d
        else:
            if any(self._ag._cmap):
                for a, cmap in zip(indices, self._ag._cmap[indices]):
                    for b, c, d in cmap:
                        if b > -1 and b in iset and c > -1 and c in iset \
                        and d > -1 and d in iset:
                            yield a, b, c, d
                    iset.remove(a)

    def iterCrossterms(self):
        """Yield crossterms formed by the atom.  Use :meth:`setCrossterms` for setting
        crossterms."""

        ag = self._ag
        acsi = self.getACSIndex()
        for crossterm in self._iterCrossterms():
            yield Crossterm(ag, crossterm, acsi) 

    def _iterDonors(self):
        """Yield pairs of indices for donored atoms that are within the pointer.
        Use :meth:`setDonors` for setting donors."""

        if self._ag._donors is None:
            LOGGER.warning('donors are not set, use `AtomGroup.setDonors`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 2 >= len(self):
            for a, b in self._ag._iterDonors():
                if a in iset and b in iset:
                    yield a, b
        else:
            if any(self._ag._domap):
                for a, dmap in zip(indices, self._ag._domap[indices]):
                    for b in dmap:
                        if b > -1 and b in iset:
                            yield a, b
                    iset.remove(a)

    def iterDonors(self):
        """Yield donors formed by the atom.  Use :meth:`setDonors` for setting
        donors."""

        ag = self._ag
        acsi = self.getACSIndex()
        for donor in self._iterDonors():
            yield Donor(ag, donor, acsi)

    def _iterAcceptors(self):
        """Yield pairs of indices for acceptored atoms that are within the pointer.
        Use :meth:`setAcceptors` for setting acceptors."""

        if self._ag._acceptors is None:
            LOGGER.warning('acceptors are not set, use `AtomGroup.setAcceptors`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 2 >= len(self):
            for a, b in self._ag._iterAcceptors():
                if a in iset and b in iset:
                    yield a, b
        else:
            if any(self._ag._acmap):
                for a, amap in zip(indices, self._ag._acmap[indices]):
                    for b in amap:
                        if b > -1 and b in iset:
                            yield a, b
                    iset.remove(a)

    def iterAcceptors(self):
        """Yield acceptors formed by the atom.  Use :meth:`setAcceptors` for setting
        acceptors."""

        ag = self._ag
        acsi = self.getACSIndex()
        for acceptor in self._iterAcceptors():
            yield Acceptor(ag, acceptor, acsi)

    def _iterNBExclusions(self):
        """Yield pairs of indices for nbexclusioned atoms that are within the pointer.
        Use :meth:`setNBExclusions` for setting nbexclusions."""

        if self._ag._nbexclusions is None:
            LOGGER.warning('nbexclusions are not set, use `AtomGroup.setNBExclusions`')

        indices = self._getIndices()
        iset = set(indices)
        if len(self._ag) / 2 >= len(self):
            for a, b in self._ag._iterNBExclusions():
                if a in iset and b in iset:
                    yield a, b
        else:
            if any(self._ag._nbemap):
                for a, nbemap in zip(indices, self._ag._nbemap[indices]):
                    for b in nbemap:
                        if b > -1 and b in iset:
                            yield a, b
                    iset.remove(a)

    def iterNBExclusions(self):
        """Yield nbexclusions formed by the atom.  Use :meth:`setNBExclusions` for setting
        nbexclusions."""

        ag = self._ag
        acsi = self.getACSIndex()
        for nbexclusion in self._iterNBExclusions():
            yield NBExclusion(ag, nbexclusion, acsi)



# -*- coding: utf-8 -*-
import os
import sys


sys.path.append(os.path.abspath('sphinxext'))
RTD = os.environ.get('READTHEDOCS', None) == 'True'
if RTD:
    tags.add('rtd')

extensions = ['sphinx.ext.todo',
              'sphinx.ext.autodoc',
              'sphinx.ext.doctest',
              'sphinx.ext.coverage',
              'sphinx.ext.extlinks',
              'sphinx.ext.graphviz',
              'sphinx.ext.ifconfig',
              'sphinx.ext.viewcode',
              'sphinx.ext.intersphinx',
              'sphinx.ext.inheritance_diagram',
              'matplotlib.sphinxext.mathmpl',
              'matplotlib.sphinxext.only_directives']

# extensions that are included in docs
extensions.extend(
             ['IPython.sphinxext.ipython_console_highlighting',
              'IPython.sphinxext.ipython_directive',
              'googleanalytics',])

source_suffix = '.rst'
master_doc = 'index'

project = u'ProDy'
copyright = u'2010-2014, University of Pittsburgh'


__version__ = ''
for _ in ['../prody/__init__.py', # building docs
          'ProDy/prody/__init__.py', # building complete website
          '../../ProDy/prody/__init__.py']: # building tutorial PDFs
    if os.path.isfile(_):
        with open(_) as inp:
            statement = ''.join([line for line in inp
                                 if line[:3] in ('__v', '__r')])
            exec(statement)
version, release = __version__, __release__


if release.endswith('dev'):
    import shlex
    from subprocess import Popen, PIPE
    _ = '../.git'
    if not os.path.isdir(_):
        _ = 'ProDy/.git'
    args = shlex.split('git --git-dir {} describe --tags --abbrev=0'.format(_))
    tag = Popen(args, stdout=PIPE, stderr=PIPE)
    rst_prolog = """

.. only:: html

    .. note::

        This documentation is for a development version of ProDy.
        There may be significant differences from the latest stable
        release ({}).

    """.format(tag.communicate()[0].strip())

exclude_patterns = ['_build', '_workdir']


add_module_names = False

show_authors = True

pygments_style = 'sphinx'

modindex_common_prefix = ['prody.']
doctest_global_setup = "from prody import *"

# -- Options for HTML output ---------------------------------------------------
if RTD:
    html_theme = 'default'
else:
    templates_path = ['_theme']
    html_theme = '_theme'
    html_theme_path = ['.']
    html_static_path = ['_static']
html_theme_options = {}

html_title = "ProDy"
html_favicon = '_static/favicon.ico'
html_last_updated_fmt = '%b %d, %Y'

html_index = 'index.html'

generic_sidebars = ['toolbox.html', 'releasenotes.html', 'howtocite.html']
html_sidebars = {'**': generic_sidebars}

html_copy_source = False
html_show_sourcelink = False

html_show_sphinx = True
html_show_copyright = True

extlinks = {
    'issue': ('https://github.com/prody/ProDy/issues/%s', 'issue '),
    'pdb': ('http://www.pdb.org/pdb/explore/explore.do?structureId=%s', ''),
    'wiki': ('http://en.wikipedia.org/wiki/%s', ''),
    'pfam': ('http://pfam-legacy.xfam.org/family/%s', ''),
    'pfamprotein': ('http://pfam-legacy.xfam.org/protein/%s', ''),
    'uniprot': ('http://www.uniprot.org/uniprot/%s', ''),
    'pdbhet': ('http://www.pdb.org/pdb/ligand/ligandsummary.do?hetId=%s', ''),
}


# ipython directive configuration
ipython_savefig_dir = os.path.join('_static', 'figures')


# -- Options for LaTeX output --------------------------------------------------
latex_paper_size = 'letter'
latex_font_size = '10pt'

latex_documents = [
    ('index', 'ProDy.tex',
     u'ProDy Documentation',
     u'Ahmet Bakan', 'manual')
]

latex_logo = '_static/logo.png'

latex_show_pagerefs = True
latex_show_urls = 'footnote'

latex_preamble = ''
latex_appendices = []

latex_domain_indices = True

latex_elements = {
    'classoptions': ',openany,oneside',
    'fontpkg': '\\usepackage{palatino}',
    'babel': '\\usepackage[english]{babel}'
}


autodoc_member_order = 'groupwise'
autodoc_default_flags = []
autoclass_content = 'both'
todo_include_todos = True

googleanalytics_enabled = True
googleanalytics_id = 'UA-19801227-1'

intersphinx_mapping = {
    'python': ('http://docs.python.org/', None),
    'numpy': ('http://docs.scipy.org/doc/numpy/', None),
    'scipy': ('http://docs.scipy.org/doc/scipy/reference/', None),
    'matplotlib': ('http://matplotlib.sourceforge.net/', None),
    'prodywebsite': ('http://www.bahargroup.org/prody/', None),
}

rst_epilog = u"""

.. _ProDy: http://www.bahargroup.org/prody
.. _Tutorials: http://www.bahargroup.org/prody/tutorials
.. _NMWiz: http://csb.pitt.edu/NMWiz
.. _VMD: http://www.ks.uiuc.edu/Research/vmd
.. _PDB: http://www.pdb.org
.. _pip: https://pypi.python.org/pypi/pip

.. _MDAnalysis: https://www.mdanalysis.org
.. _MDTraj: https://www.mdtraj.org
.. _pyparsing: http://pyparsing.wikispaces.com
.. _Matplotlib: http://matplotlib.org
.. _Biopython: http://biopython.org
.. _Chimera: https://www.cgl.ucsf.edu/chimera

.. _PyPI: http://pypi.python.org/pypi/ProDy
.. _GitHub: http://github.com/prody/ProDy

.. _IPython: http://ipython.org
.. _Python: http://www.python.org
.. _NumPy: http://www.numpy.org
.. _Scipy: http://www.scipy.org
.. _nose: http://nose.readthedocs.org
.. _Pandas: https://pandas.pydata.org/


.. |A2| replace:: Å\ :sup:`2`
"""



# -*- coding: utf-8 -*-
"""This module defines a class to handle segments of atoms in an atom group."""

from numpy import unique

from .subset import AtomSubset

__all__ = ['Segment']

class Segment(AtomSubset):

    """Instances of this class point to atoms with same segment names and
    are generated by :class:`.HierView` class.  Following built-in functions
    are customized for this class:

    * :func:`len` returns the number of chains in the segment.
    * :func:`iter` yields :class:`.Chain` instances.

    Indexing :class:`Segment` instances by a *chain identifier* (:func:`str`),
    e.g. ``A``, returns a :class:`.Chain`."""

    __slots__ = ['_ag', '_indices', '_hv', '_acsi', '_selstr']

    def __init__(self, ag, indices, hv, acsi=None, **kwargs):

        AtomSubset.__init__(self, ag, indices, acsi, **kwargs)
        self._hv = hv

    def __repr__(self):

        n_csets = self._ag.numCoordsets()
        if n_csets == 1:
            return ('<Segment: {0} from {1} ({2} chains, {3} atoms)>'
                    ).format(self.getSegname(), self._ag.getTitle(),
                             self.numChains(), self.numAtoms())
        elif n_csets > 1:
            return ('<Segment: {0} from {1} ({2} chains, {3} atoms; '
                    'active #{4} of {5} coordsets)>').format(
                    self.getSegname(), self._ag.getTitle(), self.numChains(),
                    self.numAtoms(), self.getACSIndex(), n_csets)
        else:
            return ('<Segment: {0} from {1} ({2} chains, {3} atoms; '
                    'no coordinates)>').format(self.getSegname(),
                    self._ag.getTitle(), self.numAtoms(), self.numChains())

    def __str__(self):

        return 'Segment {0}'.format(self.getSegname())

    def __getitem__(self, chid):

        return self.getChain(chid)

    def getSegname(self):
        """Returns segment name."""

        return self._ag._getSegnames()[self._indices[0]]

    def setSegname(self, segname):
        """Set segment name."""

        self.setSegnames(segname)

    def numChains(self):
        """Returns number of chains."""

        return len(set(self._getChindices()))

    __len__ = numChains

    def getChain(self, chid):
        """Returns chain with identifier *chid*."""

        self._hv.getChain(chid, self.getSegname())

    def iterChains(self):
        """Yield chains."""

        get = self._hv._getChain
        for index in unique(self._getChindices()):
            yield get(index)

    __iter__ = iterChains

    def getSelstr(self):
        """Returns selection string that selects atoms in this segment."""

        if self._selstr:
            return 'segname {0} and ({1})'.format(self.getSegname(),
                                                        self._selstr)
        else:
            return 'segname {0}'.format(self.getSegname())

    def getHierView(self, **kwargs):
        """Returns a hierarchical view of the this segment."""

        return HierView(self, **kwargs)


# -*- coding: utf-8 -*-
"""This module defines a routine for contact identification."""

from ..apptools import *

__all__ = ['prody_contacts']

def prody_contacts(**kwargs):
    """Identify contacts of a target structure with one or more ligands.
    Contacting atoms (or extended subset of atoms, such as residues) are
    outputted in PDB file format.

    :arg target: target PDB identifier or filename

    :arg ligand: ligand PDB identifier(s) or filename(s)

    :arg select: atom selection string for target structure

    :arg radius: contact radius (Å), default is ``4.0``

    :arg extend: output same ``'residue'``, ``'chain'``, or ``'segment'`` along
        with contacting atoms

    :arg prefix: prefix for output file, default is *target* filename

    :arg suffix: output filename suffix, default is *ligand* filename"""

    import prody
    LOGGER = prody.LOGGER

    target = prody.parsePDB(kwargs['target'])
    title = kwargs.get('prefix') or target.getTitle()
    selstr = kwargs.get('select')
    if selstr:
        target = target.select(selstr)
    contacts = prody.Contacts(target)
    suffix = kwargs.get('suffix', '_contacts')
    extend = kwargs.get('extend')
    radius = float(kwargs.get('radius', 4.0))
    ligands = kwargs.get('ligand')
    if len(ligands) > 1:
        outfn = lambda fn: title + suffix + '_' + fn + '.pdb'
    else:
        outfn = lambda fn: title + suffix + '.pdb'
    for pdb in ligands:
        ligand = prody.parsePDB(pdb)
        sel = contacts(radius, ligand)
        if sel:
            LOGGER.info('{0} atoms from {1} contact {2}.'
                        .format(len(sel), pdb, str(target)))
            if extend:
                sel = target.select('same ' + extend + ' as sel', sel=sel)
                LOGGER.info('Selection is extended to {0} atoms of the same '
                            '{1}(s).'.format(len(sel), extend))
            pdbfn = outfn(ligand.getTitle())
            LOGGER.info('Writing contacts into ' + pdbfn)
            prody.writePDB(pdbfn, sel)


def addCommand(commands):

    subparser = commands.add_parser('contacts',
        help='identify contacts between a target and ligand(s)')

    subparser.add_argument('--quiet', help="suppress info messages to stderr",
        action=Quiet, nargs=0)

    subparser.add_argument('--examples', action=UsageExample, nargs=0,
        help='show usage examples and exit')

    subparser.set_defaults(usage_example=
    """Identify contacts of a target structure with one or more ligands.

Fetch PDB structure 1zz2, save PDB files for individual ligands, and identify \
contacting residues of the target protein:

    $ prody select -o B11 "resname B11" 1zz2
    $ prody select -o BOG "resname BOG" 1zz2
    $ prody contacts -r 4.0 -t residue -s protein 1zz2 B11.pdb BOG.pdb
    """,
    test_examples=[(0,1,2)]
    )

    subparser.add_argument('-s', '--select', dest='select', type=str,
        metavar='SELSTR', help='selection string for target')

    subparser.add_argument('-r', '--radius', dest='radius', type=float,
        default=4.0, metavar='FLOAT',
        help='contact radius (default: %(default)s)')

    subparser.add_argument('-t', '--extend', dest='extend', type=str,
        metavar='STR', choices=set(['residue', 'chain', 'segment']),
        help=('output same residue, chain, or segment as contacting atoms'))

    subparser.add_argument('-p', '--prefix', dest='prefix', type=str,
        metavar='STR',
        help=('output filename prefix (default: target filename)'))

    subparser.add_argument('-x', '--suffix', dest='suffix', type=str,
        default='_contacts', metavar='STR',
        help=('output filename suffix (default: %(default)s)'))


    subparser.add_argument('target',
        help='target PDB identifier or filename')

    subparser.add_argument('ligand', nargs='+',
        help='ligand PDB identifier(s) or filename(s)')


    subparser.set_defaults(func=lambda opt: prody_contacts(**opt.__dict__))
    subparser.set_defaults(subparser=subparser)



"""Refine MSA application."""

from ..apptools import DevelApp

__all__ = ['evol_filter']

APP = DevelApp('filter', 'filter an MSA using sequence labels')

APP.setExample(
"""Filter sequences in an MSA based on label data.

Following example will filter human sequences:

  $ evol filter piwi_seed.slx HUMAN -e""", [])


APP.addArgument('msa',
    help='MSA filename to be filtered')

APP.addArgument('word',
    help='word to be compared to sequence label', nargs='+')

APP.addGroup('filter', 'filtering method (required)', True, True)
APP.addArgument('-s', '--startswith',
    dest='startswith',
    help='sequence label starts with given words',
    action='store_true',
    group='filter')

APP.addArgument('-e', '--endswith',
    dest='endswith',
    help='sequence label ends with given words',
    action='store_true',
    group='filter')

APP.addArgument('-c', '--contains',
    dest='contains',
    help='sequence label contains with given words',
    action='store_true',
    group='filter')

APP.addGroup('filter2', 'filter option')
APP.addArgument('-F', '--full-label',
    dest='filter_full',
    help='compare full label with word(s)',
    action='store_true',
    group='filter2')


APP.addGroup('output', 'output options')
APP.addArgument('-o', '--outname',
    dest='outname',
    help='output filename, default is msa filename with _refined suffix',
    type=str,
    metavar='STR',
    group='output')

APP.addArgument('-f', '--format',
    dest='format',
    type=str,
    metavar='STR',
    help='output MSA file format, default is same as input',
    group='output')

APP.addArgument('-z', '--compressed',
    dest='compressed',
    action='store_true',
    help='gzip refined MSA output',
    group='output')


def evol_filter(msa, *word, **kwargs):

    import prody
    from prody import MSAFile, writeMSA, LOGGER
    from os.path import splitext

    outname = kwargs.get('outname')
    if outname is None:
        outname, ext = splitext(msa)
        if ext.lower() == '.gz':
            outname, _ = splitext(msa)
        outname += '_filtered' + ext

    single = len(word) == 1
    if single:
        word = word[0]

    if kwargs.get('startswith', False):
        if single:
            filter = lambda label, seq, word=word: label.startswith(word)

    elif kwargs.get('endswith', False):
        if single:
            filter = lambda label, seq, word=word: label.endswith(word)

    elif kwargs.get('contains', False):
        if single:
            filter = lambda label, seq, word=word: word in label

    elif kwargs.get('equals', False):
        if single:
            filter = lambda label, seq, word=word: word == label
        else:
            filter = lambda label, seq, word=set(word): label in word
    else:
        raise TypeError('one of startswith, endswith, contains, or equals '
                        'must be specified')

    msa = MSAFile(msa, filter=filter,
                  filter_full=kwargs.get('filter_full', False))


    LOGGER.info('Filtered MSA is written in file: ' +
                writeMSA(outname, msa, **kwargs))


APP.setFunction(evol_filter)



"""This module contains unit tests for :mod:`~prody.dynamics`."""

import numpy as np
from numpy import arange
from numpy.testing import *

from prody.utilities import importDec
dec = importDec()

from prody import *
from prody import LOGGER
from prody.tests import unittest
from prody.tests.datafiles import *

LOGGER.verbosity = 'none'

ATOL = 1e-5
RTOL = 0

ATOMS = parseDatafile('2k39_ca')
COORDSETS = ATOMS.getCoordsets()

PCA_COV = parseDatafile('pca2k39_cov', symmetric=True)
PCA_EVALUES = parseDatafile('pca2k39_evalues')
PCA_EVECTORS = parseDatafile('pca2k39_vectors')

pca = PCA()
pca.buildCovariance(ATOMS)
pca.calcModes(n_modes=None)

n_modes = 20
pca_n_modes = PCA()
pca_n_modes.buildCovariance(ATOMS)
pca_n_modes.calcModes(n_modes=n_modes)

class TestPCA(unittest.TestCase):

    def setUp(self):

        self.model = PCA()
        self.buildMatrix = self.model.buildCovariance
        self.setMatrix = self.model.setCovariance
        self.getMatrix = self.model.getCovariance
        self.getExpected = pca.getCovariance

    def testBuildMatrixCoordsWrongType(self):
        """Test response to wrong type *coords* argument."""

        self.assertRaises(TypeError, self.buildMatrix, 'nogood')

    def testBuildMatrixWrongCoords(self):
        """Test response to wrong coords.dtype."""

        array = np.array([['a','a','a'] for i in range(10)])
        self.assertRaises(ValueError, self.buildMatrix, array)

    def testBuildMatrixCoordsArray(self):
        """Test output when *coords* is an array."""

        self.buildMatrix(COORDSETS)
        assert_equal(self.getMatrix(), self.getExpected(),
                     'failed to get correct matrix')

    def testSetMatrixWrongType(self):
        """Test response to wrong matrix type argument."""

        self.assertRaises(TypeError, self.setMatrix, list(np.ones((3,3))))

    def testSetMatrixWrongDim(self):
        """Test response to wrong dim *kirchhoff* argument."""

        self.assertRaises(ValueError, self.setMatrix, np.ones((3,4,3)))

    def testSetMatrixNonSquare(self):
        """Test response to non-square matrix."""

        self.assertRaises(ValueError, self.setMatrix, np.ones((3,4)))

    def testSetMatrixWrongDtype(self):
        """Test response to wrong matrix.dtype."""

        array = np.array([['a','a','a'] for i in range(3)])
        self.assertRaises(ValueError, self.setMatrix, array)

    def testSetMatrixAcceptableDtype(self):
        """Test response to acceptable matrix.dtype."""

        self.assertIsNone(self.setMatrix(np.ones((30, 30), int)),
                          'failed to set an acceptable array')


class TestPCAResults(TestPCA):

    def testEigenvalues(self):
        assert_allclose(pca.getEigvals(), PCA_EVALUES,
                        rtol=RTOL, atol=ATOL*100,
                        err_msg='failed to get correct eigenvalues')

    def testEigenvectors(self):
        _temp = np.abs(np.dot(pca.getEigvecs().T, PCA_EVECTORS))
        assert_allclose(_temp, np.eye(_temp.shape[0]), rtol=RTOL, atol=ATOL*10,
                       err_msg='failed to get correct eigenvectors')

    def testNumModesAll(self):
        assert_equal(pca._n_modes, len(PCA_EVALUES))

    def testNumModesSet(self):
        assert_equal(pca_n_modes._n_modes, n_modes)

    def testCovariance(self):
        assert_allclose(pca.getCovariance(), PCA_COV,
                        rtol=0, atol=ATOL,
                        err_msg='failed to get correct Covariance matrix')

    def testCovarianceSymmetry(self):
        cov = pca.getCovariance()
        assert_equal(cov, cov.T, 'Covariance is not symmetric')

if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""This module defines :class:`Selection` class for handling arbitrary subsets
of atom."""

from .subset import AtomSubset
from numpy import where

__all__ = ['Selection']

ellipsis = lambda s: s[:15] + '...' + s[-15:] if len(s) > 33 else s


class Selection(AtomSubset):

    """A class for accessing and manipulating attributes of selection of atoms
    in an :class:`.AtomGroup` instance.  Instances can be generated using
    :meth:`~.AtomGroup.select` method.  Following built-in functions are
    customized for this class:

    * :func:`len` returns the number of selected atoms
    * :func:`iter` yields :class:`.Atom` instances"""

    __slots__ = ['_ag', '_indices', '_acsi', '_selstr']

    def __init__(self, ag, indices, selstr, acsi=None, **kwargs):

        kwargs['selstr'] = selstr
        AtomSubset.__init__(self, ag, indices, acsi, **kwargs)

        # if self._getFlags('pdbter') is not None:
        #     selpdbter_flags = self.getFlags('pdbter')

        #     for i in ag._getSubset('pdbter'):
        #         if not i in indices:
        #             earlier_inds = where(indices < i)[0]
        #             if earlier_inds:
        #                 selpdbter_flags[earlier_inds[-1]] = True
        #             else:
        #                 # figure out right answer
        #                 selpdbter_flags[0] = True
                        
        # else:
        #     selpdbter_flags = None

        # self.setFlags('selpdbter', selpdbter_flags)

    def __repr__(self):

        n_csets = self._ag.numCoordsets()
        selstr = ellipsis(self._selstr)
        if n_csets:
            if n_csets == 1:
                return ('<Selection: {0} from {1} ({2} atoms)>'
                        ).format(repr(selstr), self._ag.getTitle(),
                                 len(self))
            else:
                return ('<Selection: {0} from {1} ({2} atoms; '
                        'active #{3} of {4} coordsets)>'
                        ).format(repr(selstr), self._ag.getTitle(), len(self),
                                 self.getACSIndex(), n_csets)
        else:
            return ('<Selection: {0} from {1} ({2} atoms; no '
                    'coordinates)>').format(repr(selstr), self._ag.getTitle(),
                                            len(self))

    def __getitem__(self, index):
        if isinstance(index, (str, tuple)):
            return self.getHierView()[index]
        else:
            try:
                index = self._indices[index]
                return self._ag[index]
            except:
                raise TypeError('invalid index')

    def __str__(self):

        return 'Selection {0}'.format(repr(ellipsis(self._selstr)))
    
    def getSelstr(self):
        """Returns selection string that selects this atom subset."""

        return self._selstr

    def getHierView(self, **kwargs):
        """Returns a hierarchical view of the atom selection."""

        return HierView(self, **kwargs)

    def update(self):
        """Update selection."""

        self._indices = SELECT.getIndices(self._ag, self._selstr)
